#!/usr/bin/env python
#
# Copyright (c) 2008-2009 Red Hat, Inc.
#
# This software is licensed to you under the GNU General Public License,
# version 2 (GPLv2). There is NO WARRANTY for this software, express or
# implied, including the implied warranties of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. You should have received a copy of GPLv2
# along with this software; if not, see
# http://www.gnu.org/licenses/old-licenses/gpl-2.0.txt.
#
# Red Hat trademarks are not licensed under GPLv2. No permission is
# granted to use or replicate Red Hat trademarks that are incorporated
# in this software or its documentation.
"""
Executes all tests.
"""
import os
import shutil
import sys
import tempfile

# Make sure we run from the source, this is tricky because the functional
# tests need to find both the location of the 'tito' executable script,
# and the internal tito code needs to know where to find our auxiliary Perl
# scripts. Adding an environment variable hack to the actual code to
# accommodate this for now.

TEST_SCRIPT_DIR = os.path.abspath(os.path.dirname(__file__))
SRC_DIR = os.path.normpath(os.path.join(TEST_SCRIPT_DIR, "src/"))
sys.path.insert(0, SRC_DIR)
SRC_BIN_DIR = os.path.abspath(os.path.join(TEST_SCRIPT_DIR, "bin/"))

os.environ['TITO_SRC_BIN_DIR'] = SRC_BIN_DIR

if __name__ == '__main__':
    import nose

    print("Using Python %s" % sys.version[0:3])
    print("Using nose %s" % nose.__version__[0:3])
    print("Running tito tests against: %s" % SRC_DIR)

    # Make sure no older test directories exist
    for dir in os.listdir(tempfile.gettempdir()):
        if dir.endswith('-titotest'):
            shutil.rmtree(os.path.join(tempfile.gettempdir(), dir))

    nose.main()



# Copyright (c) 2008-2010 Red Hat, Inc.
#
# This software is licensed to you under the GNU General Public License,
# version 2 (GPLv2). There is NO WARRANTY for this software, express or
# implied, including the implied warranties of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. You should have received a copy of GPLv2
# along with this software; if not, see
# http://www.gnu.org/licenses/old-licenses/gpl-2.0.txt.
#
# Red Hat trademarks are not licensed under GPLv2. No permission is
# granted to use or replicate Red Hat trademarks that are incorporated
# in this software or its documentation.
from __future__ import print_function

from contextlib import contextmanager
"""
Common operations.
"""
import errno
import fileinput
import glob
import os
import pickle
import re
import sys
import subprocess
import shlex
import shutil
import tempfile

from blessed import Terminal

from tito.compat import getstatusoutput
from tito.exception import RunCommandException
from tito.tar import TarFixer

DEFAULT_BUILD_DIR = "/tmp/tito"
DEFAULT_BUILDER = "builder"
DEFAULT_TAGGER = "tagger"
BUILDCONFIG_SECTION = "buildconfig"
SHA_RE = re.compile(r'\b[0-9a-f]{30,}\b')

# Define some shortcuts to fully qualified Builder classes to make things
# a little more concise for CLI users. Mock is probably the only one this
# is relevant for at this time.
BUILDER_SHORTCUTS = {
    'mock': 'tito.builder.MockBuilder',
    'mead': 'tito.builder.MeadBuilder',
}


def read_user_config():
    config = {}
    file_loc = os.path.expanduser("~/.titorc")
    try:
        f = open(file_loc)
    except:
        # File doesn't exist but that's ok because it's optional.
        return config

    for line in f.readlines():
        if line.strip() == "":
            continue
        tokens = line.split("=")
        if len(tokens) != 2:
            raise Exception("Error parsing ~/.titorc: %s" % line)
        # Remove whitespace from the values
        config[tokens[0].strip()] = tokens[1].strip()
    return config


def extract_sources(spec_file_lines):
    """
    Returns a list of sources from the given spec file.

    Some of these will be URL's, which is fine they will be ignored.
    We're really just after relative filenames that might live in the same
    location as the spec file, mostly used with NoTgzBuilder packages.
    """
    filenames = []
    source_pattern = re.compile(r'^Source\d+?:\s*(.*)')
    for line in spec_file_lines:
        match = source_pattern.match(line)
        if match:
            filenames.append(match.group(1))
    return filenames


def _out(msgs, prefix, color_func, stream=sys.stdout):
    if prefix is None:
        fmt = "%(msg)s"
    else:
        fmt = "%(prefix)s: %(msg)s"

    user_conf = read_user_config()
    if 'COLOR' in user_conf and (user_conf['COLOR'] == '0' or user_conf['COLOR'].lower() == 'false'):
        def color_func(x):
            return x

    if isinstance(msgs, list):
        first_line = msgs.pop(0)
        print(color_func(fmt % {'prefix': prefix, 'msg': first_line}), file=stream)
        for line in msgs:
            print(color_func("%s" % line), file=stream)
    else:
        print(color_func(fmt % {'prefix': prefix, 'msg': msgs}), file=stream)

    if 'DEBUG' in os.environ and callable(getattr(stream, 'flush', None)):
        stream.flush()


def error_out(error_msgs, die=True):
    """
    Print the given error message (or list of messages) and exit.
    """
    term = Terminal()
    _out(error_msgs, "ERROR", term.red, sys.stderr)
    if die:
        sys.exit(1)


def info_out(msgs):
    term = Terminal()
    _out(msgs, None, term.blue)


def warn_out(msgs):
    """
    Print the given error message (or list of messages) and exit.
    """
    term = Terminal()
    _out(msgs, "WARNING", term.yellow)


@contextmanager
def chdir(path):
    previous_dir = os.getcwd()
    os.chdir(path)
    try:
        yield
    finally:
        os.chdir(previous_dir)


def create_builder(package_name, build_tag,
        config, build_dir, user_config, args,
        builder_class=None, **kwargs):
    """
    Create (but don't run) the builder class. Builder object may be
    used by other objects without actually having run() called.
    """

    # Allow some shorter names for builders for CLI users.
    if builder_class in BUILDER_SHORTCUTS:
        builder_class = BUILDER_SHORTCUTS[builder_class]

    if builder_class is None:
        debug("---- Builder class is None")
        if config.has_option("buildconfig", "builder"):
            builder_class = get_class_by_name(config.get("buildconfig",
                "builder"))
        else:
            debug("---- Global config")
            builder_class = get_class_by_name(config.get(
                BUILDCONFIG_SECTION, DEFAULT_BUILDER))
    else:
        # We were given an explicit builder class as a str, get the actual
        # class reference:
        builder_class = get_class_by_name(builder_class)
    debug("Using builder class: %s" % builder_class)

    # Instantiate the builder:
    builder = builder_class(
        name=package_name,
        tag=build_tag,
        build_dir=build_dir,
        config=config,
        user_config=user_config,
        args=args,
        **kwargs)
    return builder


def find_file_with_extension(in_dir, suffix=None):
    """ Find the file with given extension in the current directory. """
    file_name = None
    debug("Looking for %s in %s" % (suffix, in_dir))
    for f in os.listdir(in_dir):
        if f.endswith(suffix):
            if file_name is not None:
                error_out("At least two %s files in directory: %s and %s" % (suffix, file_name, f))
            file_name = f
            debug("Using file: %s" % f)
    if file_name:
        return os.path.join(in_dir, file_name)
    return None


def find_spec_file(in_dir=None):
    """
    Find the first spec file in the current directory.

    Returns only the file name, rather than the full path.
    """
    if in_dir is None:
        in_dir = os.getcwd()
    result = find_file_with_extension(in_dir, '.spec')
    if result is None:
        error_out("Unable to locate a %s file in %s" % ('.spec', in_dir))
    return os.path.basename(result)


def find_gemspec_file(in_dir=None):
    """
    Find the first spec file in the current directory.

    Returns only the file name, rather than the full path.
    """
    if in_dir is None:
        in_dir = os.getcwd()
    result = find_file_with_extension(in_dir, '.gemspec')
    if result is None:
        error_out("Unable to locate a %s file in %s" % ('.gemspec', in_dir))
    return result


def find_spec_like_file(in_dir=None):
    if in_dir is None:
        in_dir = os.getcwd()
    extension_list = ['.spec', '.spec.tmpl']
    for ext in extension_list:
        result = find_file_with_extension(in_dir, ext)
        if result:
            return result
    else:
        error_out("Unable to locate files ending with %s in %s" % (list(extension_list), in_dir))


def find_cheetah_template_file(in_dir=None):
    if in_dir is None:
        in_dir = os.getcwd()
    result = find_file_with_extension(in_dir, '.spec.tmpl')
    if result is None:
        error_out("Unable to locate a %s file in %s" % ('.spec.tmpl', in_dir))
    return result


def find_mead_chain_file(in_dir=None):
    if in_dir is None:
        in_dir = os.getcwd()
    result = find_file_with_extension(in_dir, '.chain')
    if result is None:
        error_out("Unable to locate a %s file in %s" % ('.chain', in_dir))
    return result


def find_git_root():
    """
    Find the top-level directory for this git repository.

    Returned as a full path.
    """
    (status, cdup) = getstatusoutput("git rev-parse --show-cdup")
    if status > 0:
        error_out(["%s does not appear to be within a git checkout." %
                os.getcwd()])

    if cdup.strip() == "":
        cdup = "./"
    return os.path.abspath(cdup)


def tito_config_dir():
    """ Returns "rel-eng" for old tito projects and ".tito" for
    recent projects.
    """
    tito_dir = os.path.join(find_git_root(), ".tito")
    if os.path.isdir(tito_dir):
        return ".tito"
    else:
        return "rel-eng"


def extract_sha1(output):
    match = SHA_RE.search(output)
    if match:
        return match.group(0)
    else:
        return ""


def run_command(command, print_on_success=False):
    """
    Run command.
    If command fails, print status code and command output.
    """
    (status, output) = getstatusoutput(command)
    if status > 0:
        msgs = [
            "Error running command: %s\n" % command,
            "Status code: %s\n" % status,
            "Command output: %s\n" % output,
        ]
        error_out(msgs, die=False)
        raise RunCommandException(command, status, output)
    elif print_on_success:
        print("Command: %s\n" % command)
        print("Status code: %s\n" % status)
        print("Command output: %s\n" % output)
    else:
        debug("Command: %s" % command)
        debug("Status code: %s" % status)
        debug("Command output: %s\n" % output)

    return output


def run_command_print(command, print_on_success=False):
    """
    Simliar to run_command but prints each line of output on the fly.
    """
    output = []
    env = os.environ.copy()
    env['LC_ALL'] = 'C'
    p = None
    try:
        p = subprocess.Popen(command,
                             stdout=subprocess.PIPE, stderr=subprocess.STDOUT, env=env,
                             universal_newlines=True, shell=True)
    except OSError as e:
        status = e.errno
        output = e.strerror
        msgs = [
            "Error to run command: %s\n" % command,
            "Error code: %s\n" % e.errno,
            "Error description: %s\n" % e.strerror,
        ]
        error_out(msgs, die=False)
        raise RunCommandException(command, status, "\n".join(output))
    for line in run_subprocess(p):
        line = line.rstrip('\n')
        print(line)
        output.append(line)
    print("\n")
    status = p.poll()
    if status > 0:
        msgs = [
            "Error running command: %s\n" % command,
            "Status code: %s\n" % status,
            "Command output: %s\n" % output,
        ]
        error_out(msgs, die=False)
        raise RunCommandException(command, status, "\n".join(output))
    elif print_on_success:
        print("Command: %s\n" % command)
        print("Status code: %s\n" % status)
        print("Command output: %s\n" % output)
    else:
        debug("Command: %s\n" % command)
        debug("Status code: %s\n" % status)
        debug("Command output: %s\n" % output)
    return '\n'.join(output)


def run_subprocess(p):
    while True:
        retcode = p.poll()
        line = p.stdout.readline()
        if len(line) > 0:
            yield line
        if retcode is not None:
            break


def render_cheetah(template_file, destination_directory, cheetah_input):
    """Cheetah doesn't exist for Python 3, but it's the templating engine
    that Mead uses.  Instead of importing the potentially incompatible code,
    we use a command-line utility that Cheetah provides.  Yes, this is a total
    hack."""
    pickle_file = tempfile.NamedTemporaryFile(dir=destination_directory, prefix="tito-cheetah-pickle", delete=False)
    try:
        pickle.dump(cheetah_input, pickle_file, protocol=2)
        pickle_file.close()
        run_command("cheetah fill --flat --pickle=%s --odir=%s --oext=cheetah %s" %
            (pickle_file.name, destination_directory, template_file))

        # Annoyingly Cheetah won't let you specify an empty string for a file extension
        # and most Mead templates end with ".spec.tmpl"
        rendered_files = glob.glob(os.path.join(destination_directory, "*.cheetah"))

        # Cheetah returns zero even if it doesn't find the template to render.  Thanks Cheetah.
        if not rendered_files:
            error_out("Could not find rendered file in %s for %s" % (destination_directory, template_file))

        for rendered in rendered_files:
            shutil.move(rendered, os.path.splitext(rendered)[0])
    finally:
        os.unlink(pickle_file.name)


def tag_exists_locally(tag):
    (status, output) = getstatusoutput("git tag | grep %s" % tag)
    if status > 0:
        return False
    else:
        return True


def tag_exists_remotely(tag):
    """ Returns True if the tag exists in the remote git repo. """
    try:
        get_git_repo_url()
    except:
        warn_out('remote.origin does not exist. Assuming --offline, for remote tag checking.\n')
        return False
    sha1 = get_remote_tag_sha1(tag)
    debug("sha1 = %s" % sha1)
    if sha1 == "":
        return False
    return True


def get_local_tag_sha1(tag):
    tag_sha1 = run_command(
        "git ls-remote ./. --tag %s | awk '{ print $1 ; exit }'"
        % tag)
    tag_sha1 = extract_sha1(tag_sha1)
    return tag_sha1


def head_points_to_tag(tag):
    """
    Ensure the current git head is the same commit as tag.

    For some reason the git commands we normally use to fetch SHA1 for a tag
    do not work when comparing to the HEAD SHA1. Using a different command
    for now.
    """
    debug("Checking that HEAD commit is %s" % tag)
    head_sha1 = run_command("git rev-list --max-count=1 HEAD")
    tag_sha1 = run_command("git rev-list --max-count=1 %s" % tag)
    debug("   head_sha1 = %s" % head_sha1)
    debug("   tag_sha1 = %s" % tag_sha1)
    return head_sha1 == tag_sha1


def undo_tag(tag):
    """
    Executes git commands to delete the given tag and undo the most recent
    commit. Assumes you have taken necessary precautions to ensure this is
    what you want to do.
    """
    if not is_git_state_clean():
        error_out("Cannot undo a tag when the current git state is not clean!")

    # Using --merge here as it appears to undo the changes in the commit,
    # but preserve any modified files:
    output = run_command("git tag -d %s && git reset --merge HEAD^1" % tag)
    print(output)


def is_git_state_clean():
    """
    Determines if the state of the current git repository is clean or not.
    """
    (status, _) = getstatusoutput("git diff-index --quiet HEAD")
    if status != 0:
        return False

    (status, output) = getstatusoutput("git ls-files --exclude-standard --others")
    if len(output) > 0 or status > 0:
        return False

    return True


def get_remote_tag_sha1(tag):
    """
    Get the SHA1 referenced by this git tag in the remote git repo.
    Will return "" if the git tag does not exist remotely.
    """
    # TODO: X11 forwarding messages can appear in this output, find a better way
    repo_url = get_git_repo_url()
    print("Checking for tag [%s] in git repo [%s]" % (tag, repo_url))
    cmd = "git ls-remote %s --tag %s | awk '{ print $1 ; exit }'" % \
            (repo_url, tag)
    upstream_tag_sha1 = run_command(cmd)
    upstream_tag_sha1 = extract_sha1(upstream_tag_sha1)
    return upstream_tag_sha1


def check_tag_exists(tag, offline=False):
    """
    Check that the given git tag exists in a git repository.
    """
    if not tag_exists_locally(tag):
        error_out("Tag does not exist locally: [%s]" % tag)

    if offline:
        return

    tag_sha1 = get_local_tag_sha1(tag)
    debug("Local tag SHA1: %s" % tag_sha1)

    try:
        get_git_repo_url()
    except:
        warn_out('remote.origin does not exist. Assuming --offline, for remote tag checking.\n')
        return
    upstream_tag_sha1 = get_remote_tag_sha1(tag)
    if upstream_tag_sha1 == "":
        error_out(["Tag does not exist in remote git repo: %s" % tag,
            "You must tag, then git push --follow-tags"])

    debug("Remote tag SHA1: %s" % upstream_tag_sha1)

    if upstream_tag_sha1 != tag_sha1:
        error_out("Tag %s references %s locally but %s upstream." % (tag,
            tag_sha1, upstream_tag_sha1))


def debug(text, cmd=None):
    """
    Print the text if --debug was specified.
    If cmd is specified, run the command and print its output after text.
    """
    if 'DEBUG' in os.environ:
        print(text)
        if cmd:
            run_command(cmd, True)


def get_spec_version_and_release(sourcedir, spec_file_name):
    if os.path.splitext(spec_file_name)[1] == ".tmpl":
        return scrape_version_and_release(spec_file_name)

    command = ("""rpm -q --qf '%%{version}-%%{release}\n' --define """
        """"_sourcedir %s" --define 'dist %%undefined' --specfile """
        """%s 2> /dev/null | grep -e '^$' -v | head -1""" % (sourcedir, spec_file_name))
    return run_command(command)


def search_for(file_name, *args):
    """Send in a file and regular expressions as arguments.  Returns the value of the
    matching groups (or the entire matching string if no groups are in the regex) for
    each regular expression in the same order that the expressions were provided in.
    ONLY THE FIRST MATCH IS RETURNED!

    Note that this method uses re.search and not re.match so your regexs don't need
    to match the entire line.
    """
    results = [None] * len(args)
    with open(file_name, 'r') as fh:
        for line in fh:
            for index, regex in enumerate(args):
                m = re.search(regex, line)
                if not m:
                    continue

                if results[index]:
                    warn_out("Multiple matches found for %s in %s" % (regex, file_name))
                elif m.groups():
                    results[index] = m.groups()
                else:
                    results[index] = (m.string,)

        # Get the index of every regex that didn't match
        missing_results = [i for i, x in enumerate(results) if x is None]

        if len(missing_results) > 0:
            error_out("Could not find match to %s in %s" % (map(lambda x: args[x], missing_results), file_name))

        return results


def replace_spec_release(file_name, release):
    for line in fileinput.input(file_name, inplace=True):
        m = re.match(r"(\s*Release:\s*)(.*?)\s*$", line)
        # The fileinput module redirects stdout to the file handle
        # for the file being output.
        #  See https://docs.python.org/2/library/fileinput.html#fileinput.FileInput
        if m:
            print("%s%s" % (m.group(1), release))
        else:
            print(line.rstrip('\n'))


def munge_specfile(spec_file, commit_id, commit_count, fullname=None, tgz_filename=None):
    # If making a test rpm we need to get a little crazy with the spec
    # file we're building off. (Note we are modifying a temp copy of the
    # spec) Swap out the actual release for one that includes the git
    # SHA1 we're building for our test package.
    sha = commit_id[:7]

    for line in fileinput.input(spec_file, inplace=True):
        m = re.match(r'^(\s*Release:\s*)(.+?)(%{\?dist})?\s*$', line)
        if m:
            print('%s%s.git.%s.%s%s' % (
                m.group(1),
                m.group(2),
                commit_count,
                sha,
                m.group(3) or '',
            ))
            continue

        m = re.match(r'^(\s*Source0?):\s*(.+?)$', line)
        if tgz_filename and m:
            print('%s: %s' % (m.group(1), tgz_filename))
            continue

        macro = munge_setup_macro(fullname, line)
        if macro is not None:
            print(macro)
            continue

        print(line.rstrip('\n'))


def munge_setup_macro(fullname, line):
    """
    Adjust the %setup or %autosetup line in spec file to accomodate the renamed
    test source.

    Return None if the given line is not the setup or autosetup line.
    """
    m = re.match(r'^(\s*%(?:auto)?setup)(.*?)$', line)
    if fullname and m:
        macro = m.group(1)
        setup_arg = " -n %s" % fullname

        args = m.group(2)
        args_match = re.search(r'(.*?)\s+-n\s+\S+(.*)', args)
        if args_match:
            macro += args_match.group(1)
            macro += args_match.group(2)
            macro += setup_arg
        else:
            macro += args
            macro += setup_arg

        if "%autosetup" in macro:
            args_match = re.search(r'(.+?)\s+-p[01]\s+\S+(.*)', args)
            if not args_match:
                macro = "{0} -p1".format(macro)

        return macro
    return None


def scrape_version_and_release(template_file_name):
    """Ideally, we'd let RPM report the version and release of a spec file as
    in get_spec_version_and_release.  However, when we are dealing with Cheetah
    templates for Mead, RPM won't be able to parse the template file.  We have to
    fall back to using regular expressions."""
    version, release = search_for(template_file_name, r"\s*Version:\s*(.*?)\s*$", r"\s*Release:\s*(.*?)\s*$")
    version = version[0]
    release = release[0]
    release = release.replace("%{?dist}", "")
    return "%s-%s" % (version, release)


def scl_to_rpm_option(scl, silent=None):
    """ Returns rpm option which disable or enable SC and print warning if needed """
    rpm_options = ""
    cmd = "rpm --eval '%scl'"
    output = run_command(cmd).rstrip()
    if scl:
        if (output != scl) and (output != "%scl") and not silent:
            warn_out([
                "Meta package of software collection %s installed, but --scl defines %s" % (output, scl),
                "Redefining scl macro to %s for this package." % scl
            ])
        rpm_options += " --define 'scl %s'" % scl
    else:
        if (output != "%scl") and (not silent):
            warn_out([
                "Warning: Meta package of software collection %s installed, but --scl is not present." % output,
                "Undefining scl macro for this package.",
            ])
        # can be replaced by "--undefined scl" when el6 and fc17 is retired
        rpm_options += " --eval '%undefine scl'"
    return rpm_options


def get_project_name(tag=None, scl=None):
    """
    Extract the project name from the specified tag or a spec file in the
    current working directory. Error out if neither is present.
    """
    if tag is not None:
        try:
            (package, _version, _release) = tag.rsplit('-', 2)
        except ValueError:
            try:
                (package, _version) = tag.rsplit('-', 1)
            except ValueError:
                error_out("Unable to determine project name in tag: %s" % tag)
        return package
    else:
        file_path = find_spec_like_file()
        if not os.path.exists(file_path):
            error_out("spec file: %s does not exist" % file_path)

        if os.path.splitext(file_path)[1] == ".tmpl":
            name = search_for(file_path, r"\s*Name:\s*(.*?)\s*$")[0][0]
            return name
        else:
            sourcedir = os.path.dirname(file_path)
            output = run_command(
                "rpm -q --qf '%%{name}\n' %s --specfile %s --define '_sourcedir %s' "
                "2> /dev/null | grep -e '^$' -v | head -1" %
                (scl_to_rpm_option(scl, silent=True), file_path, sourcedir))

            if not output:
                error_out(["Unable to determine project name from spec file: %s" % file_path,
                    "Try rpm -q --specfile %s" % file_path,
                    "Try rpmlint -i %s" % file_path])
            return output


def replace_version(line, new_version):
    """
    Attempts to replace common setup.py version formats in the given line,
    and return the modified line. If no version is present the line is
    returned as is.

    Looking for things like version="x.y.z" with configurable case,
    whitespace, and optional use of single/double quotes.
    """
    # Mmmmm pretty regex!
    ver_regex = re.compile(r"(\s*)(version)(\s*)(=)(\s*)(['\"])(.*)(['\"])(.*)",
            re.IGNORECASE)
    m = ver_regex.match(line)
    if m:
        result_tuple = list(m.group(1, 2, 3, 4, 5, 6))
        result_tuple.append(new_version)
        result_tuple.extend(list(m.group(8, 9)))
        new_line = "%s%s%s%s%s%s%s%s%s\n" % tuple(result_tuple)
        return new_line
    else:
        return line


def get_relative_project_dir(project_name, commit):
    """
    Return the project's sub-directory relative to the git root.

    This could be a different directory than where the project currently
    resides, so we export a copy of the project's metadata from
    .tito/packages/ at the point in time of the tag we are building.
    """
    cmd = "git show %s:%s/packages/%s" % (commit, tito_config_dir(),
            project_name)
    try:
        (status, pkg_metadata) = getstatusoutput(cmd)
    except:
        cmd = "git show %s:%s/packages/%s" % (commit, "rel-eng",
            project_name)
        (status, pkg_metadata) = getstatusoutput(cmd)
    tokens = pkg_metadata.strip().split(" ")
    debug("Got package metadata: %s" % tokens)
    if status != 0:
        return None
    return tokens[1]


def get_relative_project_dir_cwd(git_root):
    """
    Returns the patch to the project we're working with relative to the
    git root using the cwd.

    *MUST* be called before doing any os.cwd().

    i.e. java/, satellite/install/Spacewalk-setup/, etc.
    """
    current_dir = os.getcwd()
    relative = current_dir[len(git_root) + 1:] + "/"
    if relative == "/":
        relative = "./"
    return relative


def get_build_commit(tag, test=False):
    """ Return the git commit we should build. """
    if test:
        return get_latest_commit(".")
    else:
        tag_sha1 = run_command(
            "git ls-remote ./. --tag %s | awk '{ print $1 ; exit }'"
            % tag)
        tag_sha1 = extract_sha1(tag_sha1)
        commit_id = run_command('git rev-list --max-count=1 %s' % tag_sha1)
        return commit_id


def get_commit_count(tag, commit_id):
    """ Return the number of commits between the tag and commit_id"""
    # git describe returns either a tag-commitcount-gSHA1 OR
    # just the tag.
    #
    # so we need to pass in the tag as well.
    # output = run_command("git describe --match=%s %s" % (tag, commit_id))
    # if tag == output:
    #     return 0
    # else:
    #     parse the count from the output
    (status, output) = getstatusoutput(
        "git describe --match=%s %s" % (tag, commit_id))

    debug("tag - %s" % tag)
    debug("output - %s" % output)

    if status != 0:
        debug("git describe of tag %s failed (%d)" % (tag, status))
        debug("going to use number of commits from initial commit")
        (status, output) = getstatusoutput(
            "git rev-list --max-parents=0 HEAD")
        output = output.split("\n")[-1]
        if status == 0:
            # output is now inital commit
            (status, output) = getstatusoutput(
                "git rev-list %s..%s --count" % (output, commit_id))
            if status == 0:
                return output
        return 0

    if tag != output:
        # tag-commitcount-gSHA1, we want the penultimate value
        cnt = output.split("-")[-2]
        return cnt

    return 0


def get_latest_commit(path="."):
    """ Return the latest git commit for the given path. """
    commit_id = run_command("git log --pretty=format:%%H --max-count=1 %s" % path)
    return commit_id


def get_commit_timestamp(sha1_or_tag):
    """
    Get the timestamp of the git commit or tag we're building. Used to
    keep the hash the same on all .tar.gz's we generate for a particular
    version regardless of when they are generated.
    """
    output = run_command(
        "git rev-list --timestamp --max-count=1 %s | awk '{print $1}'"
        % sha1_or_tag)
    return output


def create_tgz(git_root, prefix, commit, relative_dir,
    dest_tgz):
    """
    Create a .tar.gz from a projects source in git.
    """
    os.chdir(os.path.abspath(git_root))
    timestamp = get_commit_timestamp(commit)

    # Accomodate standalone projects with specfile i root of git repo:
    relative_git_dir = "%s" % relative_dir
    if relative_git_dir in ['/', './']:
        relative_git_dir = ""

    basename = os.path.splitext(dest_tgz)[0]
    initial_tar = "%s.initial" % basename

    # command to generate a git-archive
    git_archive_cmd = 'git archive --format=tar --prefix=%s/ %s:%s --output=%s' % (
        prefix, commit, relative_git_dir, initial_tar)
    run_command(git_archive_cmd)

    # Run git-archive separately if --debug was specified.
    # This allows us to detect failure early.
    # On git < 1.7.4-rc0, `git archive ... commit:./` fails!
    debug('git-archive fails if relative dir is not in git tree',
        '%s > /dev/null' % git_archive_cmd)

    fixed_tar = "%s.tar" % basename
    fixed_tar_fh = open(fixed_tar, 'wb')
    try:
        tarfixer = TarFixer(open(initial_tar, 'rb'), fixed_tar_fh, timestamp, commit)
        tarfixer.fix()
    finally:
        fixed_tar_fh.close()

    # It's a pity we can't use Python's gzip, but it doesn't offer an equivalent of -n
    return run_command("gzip -n -c < %s > %s" % (fixed_tar, dest_tgz))


def get_git_repo_url():
    """
    Return the url of this git repo.

    Uses ~/.git/config remote origin url.
    """
    return run_command("git config remote.origin.url")


def get_git_user_info():
    """
    Return the `user.name` and `user.email` git config values.
    The `--global` parameter is not specified and therefore the returned values
    are current-working-directory specific.
    """
    try:
        name = run_command('git config --get user.name')
    except:
        warn_out('user.name in ~/.gitconfig not set.\n')
        name = 'Unknown name'
    try:
        email = run_command('git config --get user.email')
    except:
        warn_out('user.email in ~/.gitconfig not set.\n')
        email = None
    return (name, email)


def get_latest_tagged_version(package_name):
    """
    Return the latest git tag for this package in the current branch.
    Uses the info in .tito/packages/package-name.

    Returns None if file does not exist.
    """
    git_root = find_git_root()
    rel_eng_dir = os.path.join(git_root, tito_config_dir())
    file_path = "%s/packages/%s" % (rel_eng_dir, package_name)
    debug("Getting latest package info from: %s" % file_path)
    if not os.path.exists(file_path):
        return None

    output = run_command("awk '{ print $1 ; exit }' %s" % file_path)
    if output is None or output.strip() == "":
        error_out("Error looking up latest tagged version in: %s" % file_path)

    return output


def normalize_class_name(name):
    """
    Just a hack to accomodate tito config files with builder/tagger
    classes referenced in the spacewalk.releng namespace, which has
    since been renamed to just tito.
    """
    look_for = "spacewalk.releng."
    if name.startswith(look_for):
        warn_out("spacewalk.releng.* namespace in tito.props is obsolete. Use tito.* instead.\n")
        name = "%s%s" % ("tito.", name[len(look_for):])
    return name


def get_script_path(scriptname):
    """
    Hack to accomodate functional tests running from source, rather than
    requiring tito to actually be installed. This variable is only set by
    test scripts, normally we assume scripts are on PATH.
    """
    # TODO: Would be nice to get rid of this hack.
    scriptpath = scriptname  # assume on PATH by default
    if 'TITO_SRC_BIN_DIR' in os.environ:
        bin_dir = os.environ['TITO_SRC_BIN_DIR']
        scriptpath = os.path.join(bin_dir, scriptname)
    return scriptpath


# 511 is 777 in octal.  Python 2 and Python 3 disagree about the right
# way to represent octal numbers.
def mkdir_p(path, mode=511):
    try:
        os.makedirs(path, mode)
    except OSError as e:
        if e.errno == errno.EEXIST and os.path.isdir(path):
            pass
        else:
            raise


def get_class_by_name(name):
    """
    Get a Python class specified by it's fully qualified name.

    NOTE: Does not actually create an instance of the object, only returns
    a Class object.
    """
    name = normalize_class_name(name)
    # Split name into module and class name:
    tokens = name.split(".")
    class_name = tokens[-1]
    module = '.'.join(tokens[0:-1])

    debug("Importing %s" % name)
    mod = __import__(module, globals(), locals(), [class_name])
    return getattr(mod, class_name)


def increase_version(version_string):
    regex = re.compile(r"^(%.*)|(.+\.)?([0-9]+)(\..*|_.*|%.*|$)")
    match = re.match(regex, version_string)
    if match:
        matches = list(match.groups())
        # Increment the number in the third match group, if there is one
        if matches[2]:
            matches[2] = str(int(matches[2]) + 1)
        # Join everything back up, skipping match groups with None
        return "".join([x for x in matches if x])

    # If no match, return the original string
    return version_string


def reset_release(release_string):
    regex = re.compile(r"(^|\.)([.0-9]+)(\.|%|$)")
    return regex.sub(r"\g<1>1\g<3>", release_string)


def increase_zstream(release_string):
    # If we do not have zstream, create .0 and then bump the version
    regex = re.compile(r"^(.*%{\?dist})$")
    bumped_string = regex.sub(r"\g<1>.0", release_string)
    return increase_version(bumped_string)


def find_wrote_in_rpmbuild_output(output):
    """
    Parse the output from rpmbuild looking for lines beginning with
    "Wrote:". Return a list of file names for each path found.
    """
    paths = []
    look_for = "Wrote: "
    for line in output.split('\n'):
        if line.startswith(look_for):
            paths.append(line[len(look_for):])
            debug("Found wrote line: %s" % paths[-1])
    if not paths:
        error_out("Unable to locate 'Wrote: ' lines in rpmbuild output: '%s'" % output)
    return paths


def compare_version(version1, version2):
    """
    Compare two version strings, returning negative if version1 is < version2,
    zero when equal and positive when version1 > version2.
    """
    def normalize(v):
        return [int(x) for x in re.sub(r'(\.0+)*$', '', v).split(".")]
    a = normalize(version1)
    b = normalize(version2)
    return (a > b) - (a < b)



# Import our builders so they can be referenced in config as tito.builder.Class
# regardless of which submodule they're in.

# flake8: noqa

from tito.builder.main import \
    Builder, \
    NoTgzBuilder, \
    GemBuilder, \
    UpstreamBuilder, \
    SatelliteBuilder, \
    MockBuilder, \
    BrewDownloadBuilder, \
    GitAnnexBuilder, \
    GitLfsBuilder, \
    MeadBuilder

from tito.builder.submodule_aware_builder import \
    SubmoduleAwareBuilder

from tito.builder.fetch import FetchBuilder



from tito.tagger import VersionTagger


class zStreamTagger(VersionTagger):
    """
    Tagger which increments the spec file zstream number instead of version.

    Used for:
        - Red Hat Service Packs (zstream)
    """

    def _tag_release(self):
        """
        Tag a new zstream of the package. (i.e. x.y.z-r%{dist}.Z+1)
        """
        self._make_changelog()
        new_version = self._bump_version(release=True)

        self._check_tag_does_not_exist(self._get_new_tag(new_version))
        self._update_changelog(new_version)
        self._update_package_metadata(new_version)

    def release_type(self):
        """ return short string "zstream release" """
        return "zstream release"



# Copyright (c) 2008-2015 Red Hat, Inc.
#
# This software is licensed to you under the GNU General Public License,
# version 2 (GPLv2). There is NO WARRANTY for this software, express or
# implied, including the implied warranties of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. You should have received a copy of GPLv2
# along with this software; if not, see
# http://www.gnu.org/licenses/old-licenses/gpl-2.0.txt.
#
# Red Hat trademarks are not licensed under GPLv2. No permission is
# granted to use or replicate Red Hat trademarks that are incorporated
# in this software or its documentation.

import sys

from contextlib import contextmanager
from mock import patch, MagicMock
from tito.compat import PY2, StringIO


# There is not many simple options to check on what distribution this is running.
# Fortunately, we only need to check for Fedora Rawhide and EPEL6, so we can
# determine it from python version. This is compatible for all distributions.
is_rawhide = sys.version_info[:2] >= (3, 8)
is_epel6 = sys.version_info[:2] == (2, 6)


if PY2:
    builtins = "__builtin__"
    builtins_open = "__builtin__.open"
    builtins_input = "__builtin__.raw_input"
else:
    builtins = "builtins"
    builtins_open = "builtins.open"
    builtins_input = "builtins.input"


file_spec = None


class Capture(object):
    class Tee(object):
        def __init__(self, stream, silent):
            self.buf = StringIO()
            self.stream = stream
            self.silent = silent

        def write(self, data):
            self.buf.write(data)
            if not self.silent:
                self.stream.write(data)

        def getvalue(self):
            return self.buf.getvalue()

        def isatty(self):
            return False

    def __init__(self, silent=False):
        self.silent = silent

    def __enter__(self):
        self.buffs = (self.Tee(sys.stdout, self.silent), self.Tee(sys.stderr, self.silent))
        self.stdout = sys.stdout
        self.stderr = sys.stderr
        sys.stdout, sys.stderr = self.buffs
        return self

    @property
    def out(self):
        return self.buffs[0].getvalue()

    @property
    def err(self):
        return self.buffs[1].getvalue()

    def __exit__(self, exc_type, exc_value, traceback):
        sys.stdout = self.stdout
        sys.stderr = self.stderr


@contextmanager
def open_mock(content, **kwargs):
    """Mock's mock_open only supports read() and write() which is not very useful.
    This context manager adds support for getting the value of what was written out
    and for iterating through a file line by line."""

    global file_spec
    if file_spec is None:
        # set on first use
        if PY2:
            file_spec = file
        else:
            import _io
            file_spec = list(set(dir(_io.TextIOWrapper)).union(set(dir(_io.BytesIO))))

    m = MagicMock(name='open', spec=open)

    handle = MagicMock(spec=file_spec)
    handle.__enter__.return_value = handle
    m.return_value = handle

    content_out = StringIO()

    with patch(builtins_open, m, create=True, **kwargs) as mo:
        stream = StringIO(content)
        rv = mo.return_value
        rv.write = lambda x: content_out.write(bytes(x, "utf-8"))
        rv.content_out = lambda: content_out.getvalue()
        rv.__iter__.return_value = iter(stream.readlines())
        rv.read.return_value = stream.read()
        yield rv



# Copyright (c) 2008-2010 Red Hat, Inc.
#
# This software is licensed to you under the GNU General Public License,
# version 2 (GPLv2). There is NO WARRANTY for this software, express or
# implied, including the implied warranties of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. You should have received a copy of GPLv2
# along with this software; if not, see
# http://www.gnu.org/licenses/old-licenses/gpl-2.0.txt.
#
# Red Hat trademarks are not licensed under GPLv2. No permission is
# granted to use or replicate Red Hat trademarks that are incorporated
# in this software or its documentation.

"""
Tito Exceptions
"""


class TitoException(Exception):
    """
    Base Tito exception.

    Does nothing but indicate that this is a custom Tito error.
    """

    def __init__(self, message):
        self.message = message

    def __str__(self):
        return "%s" % self.message


class RunCommandException(Exception):
    """ Raised by run_command() """
    def __init__(self, command, status, output):
        Exception.__init__(self, "Error running command: %s" % command)
        self.command = command
        self.status = status
        self.output = output


class ConfigException(TitoException):
    """
    Exception thrown when the user has specified invalid configuration.
    """
    pass



#
# Copyright (c) 2008-2014 Red Hat, Inc.
#
# This software is licensed to you under the GNU General Public License,
# version 2 (GPLv2). There is NO WARRANTY for this software, express or
# implied, including the implied warranties of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. You should have received a copy of GPLv2
# along with this software; if not, see
# http://www.gnu.org/licenses/old-licenses/gpl-2.0.txt.
#
# Red Hat trademarks are not licensed under GPLv2. No permission is
# granted to use or replicate Red Hat trademarks that are incorporated
# in this software or its documentation.
"""
Functional Tests for the GitAnnexBuilder.
"""

import os
import glob
import tempfile
import sys
import shutil
from nose.plugins.skip import SkipTest
from os.path import join

from functional.fixture import TitoGitTestFixture, tito

from tito.compat import *  # NOQA
from tito.common import run_command
from tito.builder import GitAnnexBuilder

PKG_NAME = "extsrc"
GIT_ANNEX_MINIMUM_VERSION = '3.20130207'


class GitAnnexBuilderTests(TitoGitTestFixture):

    def setUp(self):
        TitoGitTestFixture.setUp(self)

        # Guess based on python version.
        # Do not use anything based on uname in case we are in container.
        # Do not use `lsb_release` to avoid dependencies.
        if sys.version[0:3] == '2.4':
            raise SkipTest('git-annex is not available in epel-5')

        status, ga_version = getstatusoutput('rpm -q git-annex')
        if status != 0:
            raise SkipTest("git-annex is missing")

        # Setup test config:
        self.config = RawConfigParser()
        self.config.add_section("buildconfig")
        self.config.set("buildconfig", "builder",
                "tito.builder.GitAnnexBuilder")
        self.config.set("buildconfig", "offline",
                "true")

        os.chdir(self.repo_dir)
        specname = "extsrc-3.spec" if sys.version_info[0] == 3 else "extsrc-2.spec"
        spec = join(os.path.dirname(__file__), "specs", specname)
        self.create_project_from_spec(PKG_NAME, self.config,
                spec=spec)
        self.source_filename = 'extsrc-0.0.2.tar.gz'

        # Make a fake source file, do we need something more real?
        run_command('touch %s' % self.source_filename)
        print(run_command('git-annex init'))

        self.output_dir = tempfile.mkdtemp("-titotestoutput")

    def tearDown(self):
        run_command('chmod -R u+rw %s' % self.output_dir)
        shutil.rmtree(self.output_dir)
        TitoGitTestFixture.tearDown(self)

    def test_simple_build(self):
        run_command('git annex add %s' % self.source_filename)
        run_command('git commit -a -m "Add source."')
        # This will create 0.0.2:
        tito('tag --debug --accept-auto-changelog')
        builder = GitAnnexBuilder(PKG_NAME, None, self.output_dir,
            self.config, {}, {}, **{'offline': True})
        builder.rpm()
        self.assertEquals(1, len(list(builder.sources)))

        self.assertEquals(2, len(builder.artifacts))
        self.assertEquals(1, len(glob.glob(join(self.output_dir,
            "extsrc-0.0.2-1.*src.rpm"))))
        self.assertEquals(1, len(glob.glob(join(self.output_dir, 'noarch',
            "extsrc-0.0.2-1.*.noarch.rpm"))))
        builder.cleanup()

    def test_lock_force_supported(self):
        tito('tag --debug --accept-auto-changelog')
        builder = GitAnnexBuilder(PKG_NAME, None, self.output_dir,
            self.config, {}, {}, **{'offline': True})

        self.assertTrue(builder._lock_force_supported('5.20140107'))
        self.assertTrue(builder._lock_force_supported('5.20131213'))
        self.assertFalse(builder._lock_force_supported('5.20131127.1'))
        self.assertFalse(builder._lock_force_supported('3.20120522'))



import unittest
from tito.bugtracker import BugzillaExtractor
from mock import Mock, patch


class ExtractBugzillasTest(unittest.TestCase):

    def test_single_line(self):
        commit_log = "- 123456: Did something interesting."
        extractor = BugzillaExtractor(commit_log)
        results = extractor.extract()
        self.assertEquals(1, len(results))
        self.assertEquals("Resolves: #123456 - Did something interesting.",
                results[0])

    def test_single_with_dash(self):
        commit_log = "- 123456 - Did something interesting."
        extractor = BugzillaExtractor(commit_log)
        results = extractor.extract()
        self.assertEquals(1, len(results))
        self.assertEquals("Resolves: #123456 - Did something interesting.",
                results[0])

    def test_single_with_no_spaces(self):
        commit_log = "- 123456-Did something interesting."
        extractor = BugzillaExtractor(commit_log)
        results = extractor.extract()
        self.assertEquals(1, len(results))
        self.assertEquals("Resolves: #123456 - Did something interesting.",
                results[0])

    def test_diff_format(self):
        commit_log = "+- 123456: Did something interesting."
        extractor = BugzillaExtractor(commit_log)
        results = extractor.extract()
        self.assertEquals(1, len(results))
        self.assertEquals("Resolves: #123456 - Did something interesting.",
                results[0])

    def test_single_line_no_bz(self):
        commit_log = "- Did something interesting."
        extractor = BugzillaExtractor(commit_log)
        results = extractor.extract()
        self.assertEquals(0, len(results))

    def test_multi_line(self):
        commit_log = "- 123456: Did something interesting.\n- Another commit.\n" \
            "- 456789: A third commit."
        extractor = BugzillaExtractor(commit_log)
        results = extractor.extract()
        self.assertEquals(2, len(results))
        self.assertEquals("Resolves: #123456 - Did something interesting.",
                results[0])
        self.assertEquals("Resolves: #456789 - A third commit.",
                results[1])

    def test_single_required_flag_found(self):

        extractor = BugzillaExtractor("", required_flags=[
            'myos-1.0+', 'pm_ack+'])
        bug1 = ('123456', 'Did something interesting.')
        extractor._extract_bzs = Mock(return_value=[
            bug1])
        extractor._check_for_bugzilla_creds = Mock()

        extractor._load_bug = Mock(
            return_value=MockBug(bug1[0], ['myos-1.0+', 'pm_ack+']))

        results = extractor.extract()

        self.assertEquals(1, len(extractor.bzs))
        self.assertEquals(bug1[0], extractor.bzs[0][0])
        self.assertEquals(bug1[1], extractor.bzs[0][1])

        self.assertEquals(1, len(results))
        self.assertEquals("Resolves: #123456 - Did something interesting.",
                results[0])

    def test_required_flags_found(self):

        extractor = BugzillaExtractor("", required_flags=[
            'myos-1.0+', 'pm_ack+'])
        bug1 = ('123456', 'Did something interesting.')
        bug2 = ('444555', 'Something else.')
        bug3 = ('987654', 'Such amaze!')
        extractor._extract_bzs = Mock(return_value=[
            bug1, bug2, bug3])
        extractor._check_for_bugzilla_creds = Mock()

        bug_mocks = [
            MockBug(bug1[0], ['myos-1.0+', 'pm_ack+']),
            MockBug(bug2[0], ['myos-2.0?', 'pm_ack?']),
            MockBug(bug3[0], ['myos-1.0+', 'pm_ack+'])]

        def next_bug(*args):
            return bug_mocks.pop(0)

        extractor._load_bug = Mock(side_effect=next_bug)

        results = extractor.extract()

        self.assertEquals(2, len(extractor.bzs))
        self.assertEquals(bug1[0], extractor.bzs[0][0])
        self.assertEquals(bug1[1], extractor.bzs[0][1])
        self.assertEquals(bug3[0], extractor.bzs[1][0])
        self.assertEquals(bug3[1], extractor.bzs[1][1])

        self.assertEquals(2, len(results))
        self.assertEquals("Resolves: #123456 - Did something interesting.",
                results[0])
        self.assertEquals("Resolves: #987654 - Such amaze!",
                results[1])

    @patch("tito.bugtracker.error_out")
    def test_required_flags_missing(self, mock_error):
        required_flags = ['myos-2.0+']
        extractor = BugzillaExtractor("", required_flags)
        bug1 = ('123456', 'Did something interesting.')
        bug2 = ('444555', 'Something else.')
        bug3 = ('987654', 'Such amaze!')
        extractor._extract_bzs = Mock(return_value=[
            bug1, bug2, bug3])
        extractor._check_for_bugzilla_creds = Mock()

        bug_mocks = [
            MockBug(bug1[0], ['myos-1.0+', 'pm_ack+']),
            MockBug(bug2[0], ['myos-2.0?', 'pm_ack?']),
            MockBug(bug3[0], ['myos-1.0+', 'pm_ack+'])]

        def next_bug(*args):
            return bug_mocks.pop(0)

        extractor._load_bug = Mock(side_effect=next_bug)

        results = extractor.extract()

        self.assertEquals(0, len(extractor.bzs))
        self.assertEquals(0, len(results))
        mock_error.assert_called_once_with("No bugzillas found with required flags: %s" % required_flags)

    def test_required_flags_missing_with_placeholder(self):

        extractor = BugzillaExtractor("", required_flags=[
            'myos-2.0+'], placeholder_bz="54321")
        bug1 = ('123456', 'Did something interesting.')
        extractor._extract_bzs = Mock(return_value=[
            bug1])
        extractor._check_for_bugzilla_creds = Mock()

        extractor._load_bug = Mock(
            return_value=MockBug(bug1[0], ['myos-1.0+', 'pm_ack+']))

        results = extractor.extract()

        self.assertEquals(0, len(extractor.bzs))

        self.assertEquals(1, len(results))
        self.assertEquals("Related: #54321", results[0])

    def test_same_id_multiple_times(self):

        extractor = BugzillaExtractor("", required_flags=[
            'myos-1.0+', 'pm_ack+'])
        bug1 = ('123456', 'Did something interesting.')
        bug3 = ('123456', 'Oops, lets try again.')
        extractor._extract_bzs = Mock(return_value=[
            bug1, bug3])
        extractor._check_for_bugzilla_creds = Mock()

        extractor._load_bug = Mock(
            return_value=MockBug(bug1[0], ['myos-1.0+', 'pm_ack+']))

        results = extractor.extract()

        self.assertEquals(2, len(extractor.bzs))
        self.assertEquals(bug1[0], extractor.bzs[0][0])
        self.assertEquals(bug1[1], extractor.bzs[0][1])
        self.assertEquals(bug3[0], extractor.bzs[1][0])
        self.assertEquals(bug3[1], extractor.bzs[1][1])

        self.assertEquals(2, len(results))
        self.assertEquals("Resolves: #123456 - Did something interesting.",
                results[0])
        self.assertEquals("Resolves: #123456 - Oops, lets try again.",
                results[1])

    @patch("tito.bugtracker.error_out")
    def test_bug_doesnt_exist(self, mock_error):
        required_flags = ['myos-1.0+', 'pm_ack+']
        extractor = BugzillaExtractor("", required_flags)
        bug1 = ('123456', 'Did something interesting.')
        extractor._extract_bzs = Mock(return_value=[
            bug1])
        extractor._check_for_bugzilla_creds = Mock()

        from tito.compat import xmlrpclib
        extractor._load_bug = Mock(side_effect=xmlrpclib.Fault("", ""))

        results = extractor.extract()

        self.assertEquals(0, len(extractor.bzs))
        self.assertEquals(0, len(results))
        mock_error.assert_called_once_with("No bugzillas found with required flags: %s" % required_flags)


class MockBug(object):
    def __init__(self, bug_id, flags):
        self.flags = {}
        for flag in flags:
            self.flags[flag[0:-1]] = flag[-1]

    def get_flag_status(self, flag):
        if flag in self.flags:
            return self.flags[flag]
        else:
            return None



import os
import re
import tempfile
import unittest

from tito.cli import CLI
from tito.common import run_command
from tito.compat import getoutput

TEST_CARGO_TOML = """
[package]
name = "testing_pkg"
version = "0.0.1"
authors = ["Nobody <test@tito.com>"]

[dependencies]
"""

TEST_SPEC = """
Name:           hello_tito
Version:        0.1.7
Release:        1%{?dist}
Summary:        testing package

License:        Public Domain
Source0:        hello_tito-%{version}.tar.gz

%description


%prep
%autosetup


%build

%install

%files

%changelog
"""


def tito(argstring):
    """ Run Tito from source with given arguments. """
    return CLI().main(argstring.split(' '))


class CargoTagTest(unittest.TestCase):
    """
    Test 'CargoBump' class.
    """

    def setUp(self):
        self.repo_dir = tempfile.mkdtemp("-titocargotest")
        print("Testing in: %s" % self.repo_dir)
        os.chdir(self.repo_dir)

        self.full_pkg_dir = os.path.join(self.repo_dir, "hello_tito")
        run_command('mkdir -p %s' % self.full_pkg_dir)

        # Initialize the repo:
        os.chdir(self.full_pkg_dir)
        run_command('git init')

        # Next we tito init:
        tito("init")
        run_command('echo "offline = true" >> .tito/tito.props')
        run_command('git add .tito/tito.props')
        run_command("git commit -m 'set offline in tito.props'")

        # Init cargo project
        self.create_cargo_project()

        # Init RPM package
        self.create_rpm_package()

        # Run tito tag
        tito("tag --accept-auto-changelog")

    def write_file(self, path, contents):
        out_f = open(path, 'w')
        out_f.write(contents)
        out_f.close()

    def create_cargo_project(self):
        os.chdir(self.full_pkg_dir)
        self.write_file(os.path.join(self.full_pkg_dir, 'Cargo.toml'), TEST_CARGO_TOML)
        run_command('git add Cargo.toml')
        run_command("git commit -m 'add Cargo.toml'")

    def create_rpm_package(self):
        os.chdir(self.full_pkg_dir)
        self.write_file(os.path.join(self.full_pkg_dir, 'hello_tito.spec'), TEST_SPEC)
        run_command('git add hello_tito.spec')
        run_command("git commit -m 'add spec file'")

    def test_cargo_toml_tag(self):
        """
        Check that the version string in Cargo.toml file is the same as in spec file
        """
        changelog = getoutput("grep version Cargo.toml")
        re_version = re.compile(r'"(\d+\.\d+\.\d+)"')
        version = re_version.findall(changelog)
        assert version[0] == "0.1.8"

    def test_git_contains_cargo(self):
        """
        Check presence of the Cargo.toml file in the last commit (tag)
        """
        diff = getoutput("git diff HEAD^ -- Cargo.toml")
        assert diff.find("-version = \"0.0.1\"") != -1
        assert diff.find("+version = \"0.1.8\"") != -1



#
# Copyright (c) 2008-2014 Red Hat, Inc.
#
# This software is licensed to you under the GNU General Public License,
# version 2 (GPLv2). There is NO WARRANTY for this software, express or
# implied, including the implied warranties of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. You should have received a copy of GPLv2
# along with this software; if not, see
# http://www.gnu.org/licenses/old-licenses/gpl-2.0.txt.
#
# Red Hat trademarks are not licensed under GPLv2. No permission is
# granted to use or replicate Red Hat trademarks that are incorporated
# in this software or its documentation.
import os
import unittest

from tito.compat import *  # NOQA

UNIT_DIR = os.path.abspath(os.path.dirname(__file__))
REPO_DIR = os.path.join(UNIT_DIR, '..', '..')


class TitoUnitTestFixture(unittest.TestCase):
    """
    Fixture providing setup/teardown and utilities for unit tests.
    """
    def setUp(self):
        print
        print
        print("Testing in: %s" % REPO_DIR)
        print



# Copyright (c) 2008-2011 Red Hat, Inc.
#
# This software is licensed to you under the GNU General Public License,
# version 2 (GPLv2). There is NO WARRANTY for this software, express or
# implied, including the implied warranties of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. You should have received a copy of GPLv2
# along with this software; if not, see
# http://www.gnu.org/licenses/old-licenses/gpl-2.0.txt.
#
# Red Hat trademarks are not licensed under GPLv2. No permission is
# granted to use or replicate Red Hat trademarks that are incorporated
# in this software or its documentation.

from tito.exception import TitoException


class BuildTargetParser(object):
    """
    Parses build targets from a string in the format of:
        <branch1>:<target> <branch2>:<target> ...

    Each target must be associated with a valid branch.
    """

    def __init__(self, releaser_config, release_target, valid_branches):
        self.releaser_config = releaser_config
        self.release_target = release_target
        self.valid_branches = valid_branches

    def get_build_targets(self):
        build_targets = {}
        if not self.releaser_config.has_option(self.release_target, "build_targets"):
            return build_targets

        defined_build_targets = self.releaser_config.get(self.release_target,
                                                         "build_targets").split(" ")
        for build_target in defined_build_targets:
            # Ignore any empty from multiple spaces in the file.
            if not build_target:
                continue

            branch, target = self._parse_build_target(build_target)
            build_targets[branch] = target

        return build_targets

    def _parse_build_target(self, build_target):
        """ Parses a string in the format of branch:target """
        if not build_target:
            raise TitoException("Invalid build_target: %s. Format: <branch>:<target>"
                                % build_target)

        parts = build_target.split(":")
        if len(parts) != 2:
            raise TitoException("Invalid build_target: %s. Format: <branch>:<target>"
                                % build_target)
        branch = parts[0]
        if branch not in self.valid_branches:
            raise TitoException("Invalid build_target: %s. Unknown branch reference."
                                % build_target)
        target = parts[1]
        if not target:
            raise TitoException("Invalid build_target: %s. Empty target" % build_target)

        return (branch, target)



# Copyright (c) 2008-2011 Red Hat, Inc.
#
# This software is licensed to you under the GNU General Public License,
# version 2 (GPLv2). There is NO WARRANTY for this software, express or
# implied, including the implied warranties of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. You should have received a copy of GPLv2
# along with this software; if not, see
# http://www.gnu.org/licenses/old-licenses/gpl-2.0.txt.
#
# Red Hat trademarks are not licensed under GPLv2. No permission is
# granted to use or replicate Red Hat trademarks that are incorporated
# in this software or its documentation.
"""
Code for submitting builds for release.
"""

import copy
import os
import sys
import rpm

from tempfile import mkdtemp
import shutil

from tito.common import create_builder, debug, \
    run_command, get_project_name, warn_out, error_out
from tito.compat import PY2, dictionary_override
from tito.exception import TitoException
from tito.config_object import ConfigObject

# List of files to protect when syncing:
PROTECTED_BUILD_SYS_FILES = ('branch', 'Makefile', 'sources', ".git", ".gitignore", ".osc", "tito-mead-url",
                             "tests", "gating.yaml", ".packit.yaml", ".packit.yml", "packit.yaml", "packit.yml")

RSYNC_USERNAME = 'RSYNC_USERNAME'  # environment variable name


class Releaser(ConfigObject):
    """
    Parent class of all releasers.

    Can't really be used by itself, need to use one of the sub-classes.
    """
    GLOBAL_REQUIRED_CONFIG = ['releaser']
    REQUIRED_CONFIG = []
    OPTIONAL_CONFIG = []

    def __init__(self, name=None, tag=None, build_dir=None,
            config=None, user_config=None,
            target=None, releaser_config=None, no_cleanup=False,
            test=False, auto_accept=False, **kwargs):

        ConfigObject.__init__(self, config=config)
        config_builder_args = self._parse_builder_args(releaser_config, target)
        if test:
            config_builder_args['test'] = [True]  # builder must know to build from HEAD

        # Override with builder args from command line if any were given:
        if 'builder_args' in kwargs:
            # (in case of dupes, last one wins)
            self.builder_args = dictionary_override(
                config_builder_args,
                kwargs['builder_args']
            )
        else:
            self.builder_args = config_builder_args

        # While we create a builder here, we don't actually call run on it
        # unless the releaser needs to:
        self.offline = False
        if 'offline' in kwargs:
            self.offline = kwargs['offline']

        # Config for all releasers:
        self.releaser_config = releaser_config

        # The actual release target we're building:
        self.target = target

        # Use the builder from the release target, rather than the default
        # one defined for this git repo or sub-package:
        builder_class = None
        if self.releaser_config.has_option(self.target, 'builder'):
            builder_class = self.releaser_config.get(self.target, 'builder')
        self.builder = create_builder(name, tag,
                config,
                build_dir, user_config, self.builder_args,
                builder_class=builder_class, offline=self.offline)

        self.project_name = self.builder.project_name

        self.working_dir = mkdtemp(dir=self.builder.rpmbuild_basedir,
                prefix="release-%s" % self.builder.project_name)
        print("Working in: %s" % self.working_dir)

        self.dry_run = False
        self.test = test  # releaser must know to use builder designation rather than tag
        self.auto_accept = auto_accept  # don't ask for input, just go ahead
        self.no_cleanup = no_cleanup

        self._check_releaser_config()

    def _ask_yes_no(self, prompt="Y/N? ", default_auto_answer=True):
        if self.auto_accept:
            return default_auto_answer

        yes = ['y', 'yes', 'ok', 'sure']
        no = ['n', 'no', 'nah', 'nope']
        answers = yes + no

        while True:
            input_function = raw_input if PY2 else input
            answer = input_function(prompt).lower()
            if answer in answers:
                return answer in yes

    def _check_releaser_config(self):
        """
        Verify this release target has all the config options it needs.
        """
        for opt in self.GLOBAL_REQUIRED_CONFIG:
            if not self.releaser_config.has_option(self.target, opt):
                error_out("Release target '%s' missing required option '%s'" %
                    (self.target, opt))
        for opt in self.REQUIRED_CONFIG:
            if not self.releaser_config.has_option(self.target, opt):
                error_out("Release target '%s' missing required option '%s'" %
                    (self.target, opt))

        # TODO: accomodate 'builder.*' for yum releaser and we can use this:
        # for opt in self.releaser_config.options(self.target):
        #    if opt not in self.GLOBAL_REQUIRED_CONFIG and \
        #            opt not in self.REQUIRED_CONFIG and \
        #            opt not in self.OPTIONAL_CONFIG:
        #        raise TitoException(
        #                "Release target '%s' has unknown option '%s'" %
        #                (self.target, opt))

    def _parse_builder_args(self, releaser_config, target):
        """
        Any properties found in a releaser target section starting with
        "builder." are assumed to be builder arguments.

        i.e.:

        builder.mock = epel-6-x86_64

        Would indicate that we need to pass an argument "mock" to whatever
        builder is configured.
        """
        args = {}
        for opt in releaser_config.options(target):
            if opt.startswith("builder."):
                prefix, delimiter, opt_name = opt.partition('.')
                args.setdefault(opt_name, []).append(releaser_config.get(target, opt))
        debug("Parsed custom builder args: %s" % args)
        return args

    def release(self, dry_run=False, no_build=False, scratch=False):
        pass

    def cleanup(self):
        if not self.no_cleanup:
            debug("Cleaning up [%s]" % self.working_dir)
            run_command("rm -rf %s" % self.working_dir)

            if self.builder:
                self.builder.cleanup()
        else:
            warn_out("leaving %s (--no-cleanup)" % self.working_dir)

    def print_dry_run_warning(self, command_that_would_be_run_otherwise):
        print
        warn_out("Skipping command due to --dry-run: %s" %
                command_that_would_be_run_otherwise)
        print

    def _sync_files(self, files_to_copy, dest_dir):
        debug("Copying files: %s" % files_to_copy)
        debug("   to: %s" % dest_dir)
        os.chdir(dest_dir)

        # Need a list of just the filenames for a set comparison later:
        filenames_to_copy = []
        for filename in files_to_copy:
            filenames_to_copy.append(os.path.basename(filename))

        # Base filename for entirely new files:
        new_files = []

        # Base filenames for pre-existing files we copied over:
        copied_files = []

        # Base filenames that need to be removed by the caller:
        old_files = []

        for copy_me in files_to_copy:
            base_filename = os.path.basename(copy_me)
            dest_path = os.path.join(dest_dir, base_filename)

            if not os.path.exists(dest_path):
                print("   adding: %s" % base_filename)
                new_files.append(base_filename)
            else:
                print("   copying: %s" % base_filename)
                copied_files.append(base_filename)

            cmd = "cp %s %s" % (copy_me, dest_path)
            run_command(cmd)

        # Track filenames that will need to be deleted by the caller.
        for filename in os.listdir(dest_dir):
            if filename not in PROTECTED_BUILD_SYS_FILES and \
                    filename not in filenames_to_copy:
                print("   deleting: %s" % filename)
                old_files.append(filename)

        return new_files, copied_files, old_files


class RsyncReleaser(Releaser):
    """
    A releaser which will rsync from a remote host, build the desired packages,
    plug them in, and upload to server.

    Building of the packages is done via mock.

    WARNING: This will not work in all
    situations, depending on the current OS, and the mock target you
    are attempting to use.
    """
    REQUIRED_CONFIG = ['rsync', 'builder', 'srpm_disttag']

    # Default list of packages to copy
    filetypes = ['rpm', 'srpm', 'tgz']

    # By default run rsync with these paramaters
    rsync_args = "-rlvz"

    def __init__(self, name=None, tag=None, build_dir=None,
            config=None, user_config=None,
            target=None, releaser_config=None, no_cleanup=False,
            test=False, auto_accept=False,
            prefix="temp_dir=", **kwargs):
        Releaser.__init__(self, name, tag, build_dir, config,
                user_config, target, releaser_config, no_cleanup, test,
                auto_accept, **kwargs)

        self.build_dir = build_dir
        self.prefix = prefix

        if self.releaser_config.has_option(self.target, "scl"):
            warn_out("please rename 'scl' to 'builder.scl' in releasers.conf")
            self.builder.scl = self.releaser_config.get(self.target, "scl")

    def release(self, dry_run=False, no_build=False, scratch=False):
        self.dry_run = dry_run

        # Should this run?
        self.builder.no_cleanup = self.no_cleanup
        self.builder.tgz()

        # Check if the releaser specifies a srpm disttag:
        srpm_disttag = None
        if self.releaser_config.has_option(self.target, "srpm_disttag"):
            srpm_disttag = self.releaser_config.get(self.target, "srpm_disttag")
        self.builder.srpm(dist=srpm_disttag)

        self.builder.rpm()
        self.builder.cleanup()

        if self.releaser_config.has_option(self.target, 'rsync_args'):
            self.rsync_args = self.releaser_config.get(self.target, 'rsync_args')

        rsync_locations = self.releaser_config.get(self.target, 'rsync').split(" ")
        for rsync_location in rsync_locations:
            if RSYNC_USERNAME in os.environ:
                print("%s set, using rsync username: %s" % (RSYNC_USERNAME,
                        os.environ[RSYNC_USERNAME]))
                rsync_location = "%s@%s" % (os.environ[RSYNC_USERNAME], rsync_location)

            # Make a temp directory to sync the existing repo contents into:
            temp_dir = mkdtemp(dir=self.build_dir, prefix=self.prefix)

            self._rsync_from_remote(self.rsync_args, rsync_location, temp_dir)
            self._copy_files_to_temp_dir(temp_dir)
            self.process_packages(temp_dir)
            self.rsync_to_remote(self.rsync_args, temp_dir, rsync_location)

    def _rsync_from_remote(self, rsync_args, rsync_location, temp_dir):
        os.chdir(temp_dir)
        print("rsync %s %s %s" % (rsync_args, rsync_location, temp_dir))
        output = run_command("rsync %s %s %s" % (rsync_args, rsync_location, temp_dir))
        debug(output)

    def rsync_to_remote(self, rsync_args, temp_dir, rsync_location):
        print("rsync %s --delete %s/ %s" % (rsync_args, temp_dir, rsync_location))
        os.chdir(temp_dir)
        # TODO: configurable rsync options?
        cmd = "rsync %s --delete %s/ %s" % (rsync_args, temp_dir, rsync_location)
        if self.dry_run:
            self.print_dry_run_warning(cmd)
        else:
            output = run_command(cmd)
            debug(output)
        if not self.no_cleanup:
            debug("Cleaning up [%s]" % temp_dir)
            os.chdir("/")
            shutil.rmtree(temp_dir)
        else:
            warn_out("leaving %s (--no-cleanup)" % temp_dir)

    def _copy_files_to_temp_dir(self, temp_dir):
        os.chdir(temp_dir)

        # overwrite default self.filetypes if filetypes option is specified in config
        if self.releaser_config.has_option(self.target, 'filetypes'):
            self.filetypes = self.releaser_config.get(self.target, 'filetypes').split(" ")

        for artifact in self.builder.artifacts:
            if artifact.endswith('.tar.gz'):
                artifact_type = 'tgz'
            elif artifact.endswith('src.rpm'):
                artifact_type = 'srpm'
            elif artifact.endswith('.rpm'):
                artifact_type = 'rpm'
            else:
                continue

            if artifact_type in self.filetypes:
                print("copy: %s > %s" % (artifact, temp_dir))
                shutil.copy(artifact, temp_dir)

    def process_packages(self, temp_dir):
        """ no-op. This will be overloaded by a subclass if needed. """
        pass

    def cleanup(self):
        """ No-op, we clean up during self.release() """
        pass


class YumRepoReleaser(RsyncReleaser):
    """
    A releaser which will rsync down a yum repo, build the desired packages,
    plug them in, update the repodata, and push the yum repo back out.

    Building of the packages is done via mock.

    WARNING: This will not work in all
    situations, depending on the current OS, and the mock target you
    are attempting to use.
    """

    # Default list of packages to copy
    filetypes = ['rpm']

    # By default run createrepo_c without any paramaters
    createrepo_command = "createrepo_c ."

    def __init__(self, name=None, tag=None, build_dir=None,
            config=None, user_config=None,
            target=None, releaser_config=None, no_cleanup=False,
            test=False, auto_accept=False, **kwargs):
        RsyncReleaser.__init__(self, name, tag, build_dir, config,
                user_config, target, releaser_config, no_cleanup, test, auto_accept,
                prefix="yumrepo-", **kwargs)

    def _read_rpm_header(self, ts, new_rpm_path):
        """
        Read RPM header for the given file.
        """
        fd = os.open(new_rpm_path, os.O_RDONLY)
        header = ts.hdrFromFdno(fd)
        os.close(fd)
        return header

    def process_packages(self, temp_dir):
        self.prune_other_versions(temp_dir)
        print("Refreshing yum repodata...")
        if self.releaser_config.has_option(self.target, 'createrepo_command'):
            self.createrepo_command = self.releaser_config.get(self.target, 'createrepo_command')
        os.chdir(temp_dir)
        output = run_command(self.createrepo_command)
        debug(output)

    def prune_other_versions(self, temp_dir):
        """
        Cleanout any other version of the package we just built.

        Both older and newer packages will be removed (can be used
        to downgrade the contents of a yum repo).
        """
        os.chdir(temp_dir)
        rpm_ts = rpm.TransactionSet()
        self.new_rpm_dep_sets = {}
        for artifact in self.builder.artifacts:
            if artifact.endswith(".rpm") and not artifact.endswith(".src.rpm"):
                try:
                    header = self._read_rpm_header(rpm_ts, artifact)
                except rpm.error:
                    continue
                rpm_ds = rpm.ds(header, rpm.RPMTAG_NEVR)
                self.new_rpm_dep_sets[header['name']] = rpm_ds

        # Now cleanout any other version of the package we just built,
        # both older or newer. (can be used to downgrade the contents
        # of a yum repo)
        for filename in os.listdir(temp_dir):
            if not filename.endswith(".rpm"):
                continue
            full_path = os.path.join(temp_dir, filename)
            try:
                hdr = self._read_rpm_header(rpm_ts, full_path)
            except rpm.error:
                e = sys.exc_info()[1]
                print("error reading rpm header in '%s': %s" % (full_path, e))
                continue
            if hdr['name'] in self.new_rpm_dep_sets:
                dep_set = rpm.ds(header, rpm.RPMTAG_NEVR)
                if dep_set.EVR() < self.new_rpm_dep_sets[hdr['name']].EVR():
                    print("Deleting old package: %s" % filename)
                    run_command("rm %s" % os.path.join(temp_dir,
                        filename))


class KojiReleaser(Releaser):
    """
    Releaser for the Koji build system.

    WARNING: this is more of use to people running their own Koji instance.
    Fedora projects will most likely want to use the Fedora git releaser.
    """

    REQUIRED_CONFIG = ['autobuild_tags']
    NAME = "Koji"
    DEFAULT_KOJI_OPTS = "build --nowait"

    def __init__(self, name=None, tag=None, build_dir=None,
            config=None, user_config=None,
            target=None, releaser_config=None, no_cleanup=False,
            test=False, auto_accept=False, **kwargs):
        Releaser.__init__(self, name, tag, build_dir, config,
                user_config, target, releaser_config, no_cleanup, test, auto_accept,
                **kwargs)

        if self.releaser_config.has_option(self.target, "koji_profile"):
            self.profile = self.releaser_config.get(self.target, "koji_profile")
        else:
            self.profile = None

        if self.releaser_config.has_option(self.target, "koji_config_file"):
            self.conf_file = self.releaser_config.get(self.target, "koji_config_file")
        else:
            self.conf_file = None

        self.executable = "koji"

        self.only_tags = []
        if 'ONLY_TAGS' in os.environ:
            self.only_tags = os.environ['ONLY_TAGS'].split(' ')

        self.skip_srpm = False

    def release(self, dry_run=False, no_build=False, scratch=False):
        self.dry_run = dry_run
        self.scratch = scratch

        self._koji_release()

    def autobuild_tags(self):
        """ will return list of tags for which we are building """
        result = self.releaser_config.get(self.target, "autobuild_tags")
        return result.strip().split(" ")

    def _koji_release(self):
        """
        Lookup autobuild Koji tags from global config, create srpms with
        appropriate disttags, and submit builds to Koji.
        """
        koji_tags = self.autobuild_tags()
        print("Building release in %s..." % self.NAME)
        debug("%s tags: %s" % (self.NAME, koji_tags))

        koji_opts = self.DEFAULT_KOJI_OPTS
        if 'KOJI_OPTIONS' in self.builder.user_config:
            koji_opts = self.builder.user_config['KOJI_OPTIONS']

        scratch = self.scratch or ('SCRATCH' in os.environ and os.environ['SCRATCH'] == '1')
        if scratch:
            koji_opts = ' '.join([koji_opts, '--scratch'])

        if scratch and (self.test or self.builder.test):
            koji_opts = ' '.join([koji_opts, '--no-rebuild-srpm'])

        if self.profile:
            koji_opts = ' '.join(['--profile', self.profile, koji_opts])

        if self.conf_file:
            koji_opts = ' '.join(['--config', self.conf_file, koji_opts])

        # TODO: need to re-do this metaphor to use release targets instead:
        for koji_tag in koji_tags:
            if self.only_tags and koji_tag not in self.only_tags:
                continue
            scl = None
            if self.builder.config.has_option(koji_tag, "scl"):
                scl = self.builder.config.get(koji_tag, "scl")
            # Lookup the disttag configured for this Koji tag:
            if self.builder.config.has_option(koji_tag, "disttag"):
                disttag = self.builder.config.get(koji_tag, "disttag")
            else:
                disttag = ''
            if self.builder.config.has_option(koji_tag, "whitelist"):
                # whitelist implies only those packages can be built to the
                # tag,regardless if blacklist is also defined.
                if not self.__is_whitelisted(koji_tag, scl):
                    warn_out([
                        "%s not specified in whitelist for %s" % (self.project_name, koji_tag),
                        "   Package *NOT* submitted to %s." % self.NAME,
                    ])
                    continue
            elif self.__is_blacklisted(koji_tag, scl):
                warn_out([
                    "%s specified in blacklist for %s" % (self.project_name, koji_tag),
                    "   Package *NOT* submitted to %s." % self.NAME,
                ])
                continue

            # Getting tricky here, normally Builder's are only used to
            # create one rpm and then exit. Here we're going to try
            # to run multiple srpm builds:
            builder = self.builder
            if not self.skip_srpm:
                if scl:
                    builder = copy.copy(self.builder)
                    builder.scl = scl
                builder.srpm(dist=disttag)

            self._submit_build(self.executable, koji_opts, koji_tag, builder.srpm_location)

    def __is_whitelisted(self, koji_tag, scl):
        """ Return true if package is whitelisted in tito.props"""
        return self.builder.config.has_option(koji_tag, "whitelist") and \
            get_project_name(self.builder.build_tag, scl) in self.builder.config.get(koji_tag,
                        "whitelist").strip().split()

    def __is_blacklisted(self, koji_tag, scl):
        """ Return true if package is blacklisted in tito.props"""
        return self.builder.config.has_option(koji_tag, "blacklist") and \
            get_project_name(self.builder.build_tag, scl) in self.builder.config.get(koji_tag,
                        "blacklist").strip().split()

    def _submit_build(self, executable, koji_opts, tag, srpm_location):
        """ Submit srpm to brew/koji. """
        cmd = "%s %s %s %s" % (executable, koji_opts, tag, srpm_location)
        print("\nSubmitting build with: %s" % cmd)

        if self.dry_run:
            self.print_dry_run_warning(cmd)
            return

        output = run_command(cmd)
        print(output)


class KojiGitReleaser(KojiReleaser):
    """
    A derivative of the Koji releaser which uses a git repository to build off,
    rather than submitting srpms.
    """

    REQUIRED_CONFIG = ['autobuild_tags', 'git_url']

    def _koji_release(self):
        self.skip_srpm = True
        KojiReleaser._koji_release(self)

    def _submit_build(self, executable, koji_opts, tag, srpm_location):
        """
        Submit build to koji using the git URL from config. We will ignore
        srpm_location here.

        NOTE: overrides KojiReleaser._submit_build.
        """
        cmd = "%s %s %s %s/#%s" % \
                (executable, koji_opts, tag,
                        self.releaser_config.get(self.target, 'git_url'),
                        self.builder.build_tag)
        print("\nSubmitting build with: %s" % cmd)

        if self.dry_run:
            self.print_dry_run_warning(cmd)
            return

        output = run_command(cmd)
        print(output)



#
# Copyright (c) 2008-2014 Red Hat, Inc.
#
# This software is licensed to you under the GNU General Public License,
# version 2 (GPLv2). There is NO WARRANTY for this software, express or
# implied, including the implied warranties of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. You should have received a copy of GPLv2
# along with this software; if not, see
# http://www.gnu.org/licenses/old-licenses/gpl-2.0.txt.
#
# Red Hat trademarks are not licensed under GPLv2. No permission is
# granted to use or replicate Red Hat trademarks that are incorporated
# in this software or its documentation.
"""
Functional Tests for the FetchBuilder.
"""

import glob
import os
import sys
import shutil
import tempfile

from os.path import join

from tito.common import run_command
from tito.compat import *  # NOQA
from functional.fixture import TitoGitTestFixture, tito
from unit import Capture, is_epel6, is_rawhide

if is_epel6:
    import unittest2 as unittest
else:
    import unittest


EXT_SRC_PKG = "extsrc"

RELEASER_CONF = """
[yum-test]
releaser = tito.release.YumRepoReleaser
builder = tito.builder.FetchBuilder
srpm_disttag = .fc20
rsync = %s
"""


class FetchBuilderTests(TitoGitTestFixture):

    def setUp(self):
        TitoGitTestFixture.setUp(self)
        self.pkg_dir = join(self.repo_dir, EXT_SRC_PKG)

        specname = "extsrc-3.spec" if sys.version_info[0] == 3 else "extsrc-2.spec"
        spec = join(os.path.dirname(__file__), "specs", specname)
        shutil.copyfile(spec, os.path.join(self.repo_dir, "extsrc.spec"))
        spec = join(self.repo_dir, "extsrc.spec")

        # Setup test config:
        self.config = RawConfigParser()
        self.config.add_section("buildconfig")
        self.config.set("buildconfig", "builder",
                "tito.builder.FetchBuilder")

        self.config.add_section('builder')
        self.config.set('builder', 'fetch_strategy',
                'tito.builder.fetch.ArgSourceStrategy')

        self.create_project_from_spec(EXT_SRC_PKG, self.config,
                pkg_dir=self.pkg_dir, spec=spec)
        self.source_filename = 'extsrc-0.0.2.tar.gz'
        os.chdir(self.pkg_dir)

        # Make a fake source file, do we need something more real?
        run_command('touch %s' % self.source_filename)

        self.output_dir = tempfile.mkdtemp("-titotestoutput")

    def tearDown(self):
        TitoGitTestFixture.tearDown(self)
        # Git annex restricts permissions, change them before we remove:
        shutil.rmtree(self.output_dir)

    def test_simple_build_no_tag(self):
        # We have not tagged here. Build --rpm should just work:
        self.assertFalse(os.path.exists(
            join(self.pkg_dir, '.tito/packages/extsrc')))

        tito('build --rpm --output=%s --no-cleanup --debug --arg=source=%s ' %
                (self.output_dir, self.source_filename))
        self.assertEquals(1, len(glob.glob(join(self.output_dir,
            "extsrc-0.0.2-1.*src.rpm"))))
        self.assertEquals(1, len(glob.glob(join(self.output_dir,
            "noarch/extsrc-0.0.2-1.*noarch.rpm"))))

    def test_tag_rejected(self):
        with Capture(silent=True):
            self.assertRaises(SystemExit, tito,
                    'build --tag=extsrc-0.0.1-1 --rpm --output=%s --arg=source=%s ' %
                    (self.output_dir, self.source_filename))

    def _setup_fetchbuilder_releaser(self, yum_repo_dir):
        self.write_file(join(self.repo_dir, '.tito/releasers.conf'),
                RELEASER_CONF % yum_repo_dir)

    # createrepo_c (0.15.5+ which is in rawhide) currently coredumps
    # https://github.com/rpm-software-management/createrepo_c/issues/202
    @unittest.skipIf(is_rawhide, "Re-enable once createrepo_c #202 gets fixed")
    def test_with_releaser(self):
        yum_repo_dir = os.path.join(self.output_dir, 'yum')
        run_command('mkdir -p %s' % yum_repo_dir)
        self._setup_fetchbuilder_releaser(yum_repo_dir)
        tito('release --debug yum-test --arg source=%s' %
                self.source_filename)

        self.assertEquals(1, len(glob.glob(join(yum_repo_dir,
            "extsrc-0.0.2-1.*noarch.rpm"))))
        self.assertEquals(1, len(glob.glob(join(yum_repo_dir,
            "repodata/repomd.xml"))))



# Copyright (c) 2008-2011 Red Hat, Inc.
#
# This software is licensed to you under the GNU General Public License,
# version 2 (GPLv2). There is NO WARRANTY for this software, express or
# implied, including the implied warranties of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. You should have received a copy of GPLv2
# along with this software; if not, see
# http://www.gnu.org/licenses/old-licenses/gpl-2.0.txt.
#
# Red Hat trademarks are not licensed under GPLv2. No permission is
# granted to use or replicate Red Hat trademarks that are incorporated
# in this software or its documentation.

from tito.release import Releaser


class DummyReleaser(Releaser):

    def __init__(self, name=None, version=None, tag=None, build_dir=None,
            config=None, user_config=None):
        Releaser.__init__(self, name, version, tag, build_dir,
                config, user_config)
        pass

    def release(self, dry_run=False):
        print("DUMMY RELEASE!!!!!!!!!!!!!!!!!")



import re
import os

from tito.common import debug, run_command


class CargoBump:
    """
    Cargo is package manager for the Rust programming
    language: http://doc.crates.io/manifest.html
    It uses Cargo.toml file as its configuration file.
    XXX: I'm not including a Toml parser, because I
    don't want to introduce new dependencies.
    """

    @staticmethod
    def tag_new_version(project_path, new_version_release):
        """
        Find the line with version number  and change
        it to contain the new version.
        """
        file_name = "Cargo.toml"
        config_file = os.path.join(project_path, file_name)

        if not os.path.exists(config_file):
            debug('Cargo.toml file not found, this is probably not a Rust project')
            return

        debug("Found Cargo.toml file, attempting to update the version.")
        # We probably don't want version-release in config file as
        # release is an RPM concept
        new_version = new_version_release.split('-')[0]
        file_buffer = []

        # Read file line by line and replace version when found
        with open(config_file, 'r') as cfgfile:
            file_buffer = CargoBump.process_cargo_toml(cfgfile, new_version)

        # Write the buffer back into the file
        with open(config_file, 'w') as cfgfile:
            cfgfile.writelines(map(lambda x: x + "\n", file_buffer))

        # Add Cargo.toml into git index
        run_command("git add %s" % file_name)

    @staticmethod
    def process_cargo_toml(input_string, new_version):
        file_buffer = []
        pkg_label = re.compile(r'^\[package\]$')
        label = re.compile(r'^\[.*\]$')
        version = re.compile(r'(^version\s*=\s*)["\'](.+)["\'](.*$)')
        lines = [line.rstrip('\n') for line in input_string]
        state = 1
        for line in lines:
            # Looking for [package] label
            if state == 1:
                file_buffer.append(line)
                if re.match(pkg_label, line):
                    state = 2
            elif state == 2:
                # Looking for version = "x.x.x" line
                if re.match(version, line):
                    v = re.split(version, line)
                    file_buffer.append(v[1] + '"' + new_version + '"' + v[3])
                    state = 3
                else:
                    file_buffer.append(line)
                # if we found another label before version, it's probably not there
                if re.match(label, line):
                    state = 3
            # Just copy the rest of the file into the buffer
            else:
                file_buffer.append(line)

        return file_buffer



# Copyright (c) 2013 Red Hat, Inc.
#
# This software is licensed to you under the GNU General Public License,
# version 2 (GPLv2). There is NO WARRANTY for this software, express or
# implied, including the implied warranties of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. You should have received a copy of GPLv2
# along with this software; if not, see
# http://www.gnu.org/licenses/old-licenses/gpl-2.0.txt.
#
# Red Hat trademarks are not licensed under GPLv2. No permission is
# granted to use or replicate Red Hat trademarks that are incorporated
# in this software or its documentation.
import datetime
import os.path
import subprocess
import sys
import tempfile

try:
    # Optional dependency available on Fedora and EPEL9+
    # Without it, branch aliases in releasers.conf won't work.
    from fedora_distro_aliases import get_distro_aliases
except ImportError:
    get_distro_aliases = None

from tito.common import (
    run_command,
    debug,
    extract_sources,
    error_out,
    chdir,
    warn_out,
    info_out,
    find_mead_chain_file,
    get_git_user_info,
)
from tito.compat import getoutput, getstatusoutput, write
from tito.release import Releaser
from tito.release.main import PROTECTED_BUILD_SYS_FILES
from tito.buildparser import BuildTargetParser
from tito.exception import RunCommandException
from tito.bugtracker import BugzillaExtractor, MissingBugzillaCredsException
import getpass
from string import Template

MEAD_SCM_USERNAME = 'MEAD_SCM_USERNAME'


class FedoraGitReleaser(Releaser):

    REQUIRED_CONFIG = ['branches']

    def __init__(self, name=None, tag=None, build_dir=None,
            config=None, user_config=None,
            target=None, releaser_config=None, no_cleanup=False,
            test=False, auto_accept=False, **kwargs):
        Releaser.__init__(self, name, tag, build_dir, config,
                user_config, target, releaser_config, no_cleanup, test,
                auto_accept, **kwargs)

        if 'FEDPKG_USER' in user_config:
            self.cli_tool = "fedpkg --user=%s" % user_config["FEDPKG_USER"]
        else:
            self.cli_tool = "fedpkg"

        self.git_branches = self._branches()

        # check .tito/releasers.conf
        if self.releaser_config.has_option(self.target, "remote_git_name"):
            overwrite_checkout = self.releaser_config.get(self.target, "remote_git_name")
        # fallback to old, incorrect, location: .tito/packages/tito.props
        elif self.config.has_option(self.target, "remote_git_name"):
            overwrite_checkout = self.config.get(self.target, "remote_git_name")
        else:
            overwrite_checkout = None
        if overwrite_checkout:
            self.project_name = overwrite_checkout

        self.package_workdir = os.path.join(self.working_dir,
                self.project_name)

        build_target_parser = BuildTargetParser(self.releaser_config, self.target,
                                                self.git_branches)
        self.build_targets = build_target_parser.get_build_targets()

        # Files we should copy to git during a release:
        self.copy_extensions = (".spec", ".changes", ".rpmlintrc", ".patch")

    def release(self, dry_run=False, no_build=False, scratch=False):
        self.scratch = scratch
        self.dry_run = dry_run
        self.no_build = no_build
        self._git_release()

    def _branches(self):
        names = self.releaser_config.get(self.target, "branches").split(" ")

        # This is a bit hacky because our inheritence hierarchy is messed up.
        # RHEL, and CentOS releasers inherits from the Fedora releaser instead
        # of all of them having a common ancestor. For now, we only want the
        # support for branch aliases in the `FedoraGitReleaser`. There is no
        # reason to query Fedora PDC in RHEL and CentOS releasers.
        if self.cli_tool != "fedpkg":
            return names

        if not get_distro_aliases:
            warn_out("Missing optional dependency `fedora-distro-aliases'")
            warn_out("Branch aliases in releasers.conf won't be evaluated.")
            return names

        aliases = get_distro_aliases()
        result = []
        for name in names:
            branches = [name]
            if name in aliases:
                branches = [x.branch for x in aliases[name]]
            result.extend(branches)

        # It is probably a good idea to release into rawhide first
        if "rawhide" in result:
            result.remove("rawhide")
            result.insert(0, "rawhide")

        return result

    def _get_build_target_for_branch(self, branch):
        if branch in self.build_targets:
            return self.build_targets[branch]
        return None

    def _git_release(self):
        getoutput("mkdir -p %s" % self.working_dir)
        with chdir(self.working_dir):
            run_command("%s clone %s" % (self.cli_tool, self.project_name))

        with chdir(self.package_workdir):
            run_command("%s switch-branch %s" % (self.cli_tool, self.git_branches[0]))

        # Set git user config to the distgit clone based on the current project
        self._git_set_user_config()

        # Mead builds need to be in the git_root.  Other builders are agnostic.
        with chdir(self.git_root):
            self.builder.tgz()
            self.builder.copy_and_download_extra_sources()

        if self.test:
            self.builder._setup_test_specfile()

        self._git_sync_files(self.package_workdir)
        self._git_upload_sources(self.package_workdir)
        self._git_user_confirm_commit(self.package_workdir)

    def _git_set_user_config(self):
        fullname, email = get_git_user_info()
        email = email or ""
        with chdir(self.package_workdir):
            run_command("git config user.name '{0}'".format(fullname))
            run_command("git config user.email '{0}'".format(email))

    def _get_bz_flags(self):
        required_bz_flags = None
        if self.releaser_config.has_option(self.target,
            'required_bz_flags'):
            required_bz_flags = self.releaser_config.get(self.target,
                'required_bz_flags').split(" ")
            debug("Found required flags: %s" % required_bz_flags)

        placeholder_bz = None
        if self.releaser_config.has_option(self.target,
            'placeholder_bz'):
            placeholder_bz = self.releaser_config.get(self.target,
                'placeholder_bz')
            debug("Found placeholder bugzilla: %s" % placeholder_bz)

        return (required_bz_flags, placeholder_bz)

    def _confirm_commit_msg(self, diff_output):
        """
        Generates a commit message in a temporary file, gives the user a
        chance to edit it, and returns the filename to the caller.
        """

        fd, name = tempfile.mkstemp()
        debug("Storing commit message in temp file: %s" % name)
        write(fd, "Update %s to %s\n" % (self.project_name,
            self.builder.build_version))
        # Write out Resolves line for all bugzillas we see in commit diff:
        # TODO: move to DistGitBuilder only?
        try:
            (required_bz_flags, placeholder_bz) = self._get_bz_flags()
            extractor = BugzillaExtractor(diff_output,
                required_flags=required_bz_flags,
                placeholder_bz=placeholder_bz)
            for line in extractor.extract():
                write(fd, line + "\n")
        except MissingBugzillaCredsException:
            error_out([
                "Releaser specifies required flags but you have not configured",
                "a ~/.bugzillarc with your bugzilla credentials.",
                "Example:",
                "",
                "[bugzilla.redhat.com]",
                "user = dgoodwin@redhat.com",
                "password = mypassword"])

        print("")
        print("##### Commit message: #####")
        print("")

        os.lseek(fd, 0, 0)
        f = os.fdopen(fd)
        for line in f.readlines():
            print(line)
        f.close()

        print("")
        print("###############################")
        print("")
        if self._ask_yes_no("Would you like to edit this commit message? [y/n] ", False):
            debug("Opening editor for user to edit commit message in: %s" % name)
            editor = 'vi'
            if "EDITOR" in os.environ:
                editor = os.environ["EDITOR"]
            subprocess.call(editor.split() + [name])

        return name

    def _git_user_confirm_commit(self, project_checkout):
        """ Prompt user if they wish to proceed with commit. """
        print("")
        text = "Running 'git diff' in: %s" % project_checkout
        print("#" * len(text))
        print(text)
        print("#" * len(text))
        print("")

        main_branch = self.git_branches[0]

        os.chdir(project_checkout)

        # Newer versions of git don't seem to want --cached here? Try both:
        (unused, diff_output) = getstatusoutput("git diff --cached")
        if diff_output.strip() == "":
            debug("git diff --cached returned nothing, falling back to git diff.")
            (unused, diff_output) = getstatusoutput("git diff")

        if diff_output.strip() == "":
            print("No changes in main branch, skipping commit for: %s" % main_branch)
        else:
            print(diff_output)
            print("")
            print("##### Please review the above diff #####")
            if not self._ask_yes_no("Do you wish to proceed with commit? [y/n] "):
                print("Fine, you're on your own!")
                self.cleanup()
                sys.exit(1)

            print("Proceeding with commit.")
            commit_msg_file = self._confirm_commit_msg(diff_output)
            cmd = '%s commit -F %s' % (self.cli_tool,
                    commit_msg_file)
            debug("git commit command: %s" % cmd)
            print
            if self.dry_run:
                self.print_dry_run_warning(cmd)
            else:
                print("Proceeding with commit.")
                os.chdir(self.package_workdir)
                run_command(cmd)

            os.unlink(commit_msg_file)

        cmd = self._push_command()
        if self.dry_run:
            self.print_dry_run_warning(cmd)
        else:
            # Push
            print(cmd)
            try:
                run_command(cmd)
            except RunCommandException as e:
                error_out("`%s` failed with: %s" % (cmd, e.output))

        if not self.no_build:
            self._build(main_branch)

        for branch in self.git_branches[1:]:
            info_out("Merging branch: '%s' -> '%s'" % (main_branch, branch))
            run_command("%s switch-branch %s" % (self.cli_tool, branch))
            self._merge(main_branch)

            cmd = "git push origin %s:%s" % (branch, branch)
            if self.dry_run:
                self.print_dry_run_warning(cmd)
            else:
                print(cmd)
                try:
                    run_command(cmd)
                except RunCommandException as e:
                    error_out("`%s` failed with: %s" % (cmd, e.output))

            if not self.no_build:
                self._build(branch)

            print

    def _push_command(self):
        return "%s push" % self.cli_tool

    def _merge(self, main_branch):
        try:
            run_command("git merge %s" % main_branch)
        except:
            print
            warn_out("Conflicts occurred during merge.")
            print
            print("You are being dropped to a shell in the working directory.")
            print
            print("Please resolve this by doing the following:")
            print
            print("  1. List the conflicting files: git ls-files --unmerged")
            print("  2. Edit each resolving the conflict and then: git add FILENAME")
            print("  4. Commit the result when you are done: git commit")
            print("  4. Return to the tito release: exit")
            print
            # TODO: maybe prompt y/n here
            os.system(os.environ['SHELL'])

    def _build(self, branch):
        """ Submit a Fedora build from current directory. """
        target_param = ""
        scratch_param = ""
        build_target = self._get_build_target_for_branch(branch)
        if build_target:
            target_param = "--target %s" % build_target
        if self.scratch:
            scratch_param = "--scratch"

        build_cmd = "%s build --nowait %s %s" % (self.cli_tool, scratch_param, target_param)

        if self.dry_run:
            self.print_dry_run_warning(build_cmd)
            return

        info_out("Submitting build: %s" % build_cmd)
        (status, output) = getstatusoutput(build_cmd)
        if status > 0:
            if "already been built" in output:
                warn_out("Build has been submitted previously, continuing...")
            else:
                error_out([
                    "Unable to submit build."
                    "  Status code: %s\n" % status,
                    "  Output: %s\n" % output,
                ])

        # Print the task ID and URL:
        for line in extract_task_info(output):
            print(line)

    def _git_upload_sources(self, project_checkout):
        """
        Upload any tarballs to the lookaside directory. (if necessary)
        Uses the "fedpkg new-sources" command.
        """
        sources = [x for x in self.builder.sources
                   if not x.endswith(self.copy_extensions)]

        if not sources:
            debug("No sources need to be uploaded.")
            return

        print("Uploading sources to lookaside:")
        os.chdir(project_checkout)
        cmd = '%s new-sources %s' % (self.cli_tool, " ".join(sources))
        debug(cmd)

        if self.dry_run:
            self.print_dry_run_warning(cmd)
            return

        output = run_command(cmd)
        debug(output)
        debug("Adding write permission for:")
        for filename in sources:
            run_command("chmod u+w %s" % filename)

    def _list_files_to_copy(self):
        """
        Returns a list of the full file paths for each file that should be
        copied from our git project into the build system checkout. This
        is used to sync files to git during a release.

        i.e. spec file, .patches.

        It is assumed that any file found in the build system checkout
        but not in this list, and not in the protected files list, should
        probably be cleaned up.
        """
        # Include the spec file explicitly, in the case of SatelliteBuilder
        # we modify and then use a spec file copy from a different location.
        files_to_copy = [self.builder.spec_file]  # full paths

        f = open(self.builder.spec_file, 'r')
        lines = f.readlines()
        f.close()
        source_filenames = extract_sources(lines)
        debug("Watching for source filenames: %s" % source_filenames)

        for filename in os.listdir(self.builder.rpmbuild_gitcopy):
            full_filepath = os.path.join(self.builder.rpmbuild_gitcopy, filename)
            if os.path.isdir(full_filepath):
                # skip it
                continue
            if filename in PROTECTED_BUILD_SYS_FILES:
                debug("   skipping:  %s (protected file)" % filename)
                continue
            elif filename.endswith(".spec"):
                # Skip the spec file, we already copy this explicitly as it
                # can come from a couple different locations depending on which
                # builder is in use.
                continue

            # Check if file looks like it matches a Source line in the spec file:
            if filename in source_filenames:
                debug("   copying:   %s" % filename)
                files_to_copy.append(full_filepath)
                continue

            # Check if file ends with something this builder subclass wants
            # to copy:
            copy_it = False
            for extension in self.copy_extensions:
                if filename.endswith(extension):
                    copy_it = True
                    continue
            if copy_it:
                debug("   copying:   %s" % filename)
                files_to_copy.append(full_filepath)

        return files_to_copy

    def _git_sync_files(self, project_checkout):
        """
        Copy files from our git into each git build branch and add them.

        A list of safe files is used to protect critical files both from
        being overwritten by a git file of the same name, as well as being
        deleted after.
        """

        # Build the list of all files we will copy:
        debug("Searching for files to copy to build system git:")
        files_to_copy = self._list_files_to_copy()

        os.chdir(project_checkout)

        new, copied, old =  \
                self._sync_files(files_to_copy, project_checkout)

        os.chdir(project_checkout)

        # Git add everything:
        for add_file in (new + copied):
            run_command("git add %s" % add_file)

        # Cleanup obsolete files:
        for cleanup_file in old:
            # Can't delete via full path, must not chdir:
            run_command("git rm %s" % cleanup_file)


class DistGitReleaser(FedoraGitReleaser):
    def __init__(self, name=None, tag=None, build_dir=None,
            config=None, user_config=None,
            target=None, releaser_config=None, no_cleanup=False,
            test=False, auto_accept=False, **kwargs):
        FedoraGitReleaser.__init__(self, name, tag, build_dir, config,
                user_config, target, releaser_config, no_cleanup, test,
                auto_accept, **kwargs)
        # Override the cli_tool config from Fedora:
        if "RHPKG_USER" in user_config:
            self.cli_tool = "rhpkg --user=%s" % user_config["RHPKG_USER"]
        else:
            self.cli_tool = "rhpkg"


class CentosGitReleaser(FedoraGitReleaser):
    """
    This process follows the choreography defined for the Centos9 Stream.
    It will create a branch on the user's fork of the main repo.
    That branch is merged to the main repo via pull request on the
    web interface. There is no build option usable from this process
    because it is not synchronous.

    It is limited to only acting on a single Centos stream. Only the first
    entry in 'branches' in the conf file will be used.
    """

    def __init__(self, name=None, tag=None, build_dir=None,
            config=None, user_config=None,
            target=None, releaser_config=None, no_cleanup=False,
            test=False, auto_accept=False, **kwargs):
        FedoraGitReleaser.__init__(self, name, tag, build_dir, config,
                user_config, target, releaser_config, no_cleanup, test,
                auto_accept, **kwargs)
        # Override the cli_tool config from Fedora:
        if "CENTPKG_USER" in user_config:
            self.cli_tool = "centpkg --user={user}".format(user=user_config["CENTPKG_USER"])
            self.username = user_config["CENTPKG_USER"]
        else:
            self.cli_tool = "centpkg"
            self.username = getpass.getuser()

    def _git_release(self):
        os.makedirs(self.working_dir, exist_ok=True)
        with chdir(self.working_dir):
            run_command("{cli_tool} clone {project_name}".format(cli_tool=self.cli_tool,
                                                                 project_name=self.project_name))

        with chdir(self.package_workdir):
            run_command("{cli_tool} fork".format(cli_tool=self.cli_tool))
            run_command("git fetch {username}".format(username=self.username))
            self.new_branch_name = "release-branch-{timestamp}".format(
                timestamp=int(datetime.datetime.utcnow().timestamp()))
            run_command("git checkout -b {new_branch_name} {username}/{branch}".format(
                new_branch_name=self.new_branch_name,
                username=self.username,
                branch=self.git_branches[0]))

        # Mead builds need to be in the git_root.  Other builders are agnostic.
        with chdir(self.git_root):
            self.builder.tgz()
            self.builder.copy_and_download_extra_sources()

        if self.test:
            self.builder._setup_test_specfile()

        self._git_sync_files(self.package_workdir)
        self._git_upload_sources(self.package_workdir)
        self.no_build = True
        self.git_branches = self.git_branches[:1]
        self._git_user_confirm_commit(self.package_workdir)

    def _push_command(self):
        return "git push {username} {new_branch_name}".format(username=self.username,
                                                             new_branch_name=self.new_branch_name)


class DistGitMeadReleaser(DistGitReleaser):
    REQUIRED_CONFIG = ['mead_scm', 'branches']

    def __init__(self, name=None, tag=None, build_dir=None,
        config=None, user_config=None,
        target=None, releaser_config=None, no_cleanup=False,
        test=False, auto_accept=False,
        prefix="temp_dir=", **kwargs):

        if 'builder_args' in kwargs:
            kwargs['builder_args']['local'] = False

        DistGitReleaser.__init__(self, name, tag, build_dir, config,
                user_config, target, releaser_config, no_cleanup, test,
                auto_accept, **kwargs)

        self.mead_scm = self.releaser_config.get(self.target, "mead_scm")

        if self.releaser_config.has_option(self.target, "mead_push_url"):
            self.push_url = self.releaser_config.get(self.target, "mead_push_url")
        else:
            self.push_url = self.mead_scm

        # rhpkg maven-build takes an optional override --target:
        self.brew_target = None
        if self.releaser_config.has_option(self.target, "target"):
            self.brew_target = self.releaser_config.get(self.target, "target")

        # If the push URL contains MEAD_SCM_URL, we require the user to set this
        # in ~/.titorc before they can run this releaser. This allows us to
        # use push URLs that require username auth, but still check a generic
        # URL into source control:
        if MEAD_SCM_USERNAME in self.push_url:
            debug("Push URL contains %s, checking for value in ~/.titorc" %
                MEAD_SCM_USERNAME)
            if MEAD_SCM_USERNAME in user_config:
                user = user_config[MEAD_SCM_USERNAME]
            else:
                user = getpass.getuser()
                warn_out("You should specify MEAD_SCM_USERNAME in '~/.titorc'.  Using %s for now" % user)

            self.push_url = self.push_url.replace(MEAD_SCM_USERNAME, user)

    def _sync_mead_scm(self):
        cmd = "git push %s %s" % (self.push_url, self.builder.build_tag)

        if self.dry_run:
            self.print_dry_run_warning(cmd)
            return

        with chdir(self.git_root):
            info_out("Syncing local repo with %s" % self.push_url)
            try:
                run_command(cmd)
            except RunCommandException as e:
                if "rejected" in e.output:
                    if self._ask_yes_no("The remote rejected a push.  Force push? [y/n] ", False):
                        run_command("git push --force %s %s" % (self.mead_scm, self.builder.build_tag))
                    else:
                        error_out("Could not sync with %s" % self.mead_scm)
                raise

    def _git_release(self):
        self._sync_mead_scm()
        DistGitReleaser._git_release(self)

    def _git_upload_sources(self, project_checkout):
        chain_file = find_mead_chain_file(self.builder.rpmbuild_gitcopy)
        with open(chain_file, 'r') as f:
            template = Template(f.read())

            ref = self.builder.build_tag
            if self.test:
                ref = self.builder.git_commit_id

            values = {
                'mead_scm': self.mead_scm,
                'git_ref': ref,
                # Each property on its own line with a few leading spaces to indicate
                # that it's a continuation
                'maven_properties': "\n  ".join(self.builder.maven_properties),
                'maven_options': " ".join(self.builder.maven_args),
            }
            rendered_chain = template.safe_substitute(values)

        with chdir(project_checkout):
            with open("mead.chain", "w") as f:
                f.write(rendered_chain)

            cmd = "git add mead.chain"
            if self.dry_run:
                self.print_dry_run_warning(cmd)
                info_out("Chain file contents:\n%s" % rendered_chain)
            else:
                run_command(cmd)

    def _build(self, branch):
        """ Submit a Mead build from current directory. """
        target_param = ""
        build_target = self._get_build_target_for_branch(branch)
        if build_target:
            target_param = "--target=%s" % build_target

        build_cmd = [self.cli_tool, "maven-chain", "--nowait"]

        if self.brew_target:
            build_cmd.append("--target=%s" % self.brew_target)

        build_cmd.append("--ini=%s" % (os.path.join(self.package_workdir, "mead.chain")))
        build_cmd.append(target_param)

        if self.scratch:
            build_cmd.append("--scratch")

        build_cmd = " ".join(build_cmd)

        if self.dry_run:
            self.print_dry_run_warning(build_cmd)
            return

        info_out("Submitting build: %s" % build_cmd)
        (status, output) = getstatusoutput(build_cmd)
        if status > 0:
            if "already been built" in output:
                warn_out("Build has been submitted previously, continuing...")
            else:
                error_out([
                    "Unable to submit build.",
                    "  Status code: %s\n" % status,
                    "  Output: %s\n" % output,
                ])

        # Print the task ID and URL:
        for line in extract_task_info(output):
            print(line)


def extract_task_info(output):
    """ Extracts task ID and URL from koji/brew build output. """
    task_lines = []
    for line in output.splitlines():
        if "Created task" in line:
            task_lines.append(line)
        elif "Task info" in line:
            task_lines.append(line)
    return task_lines



# flake8: noqa

from tito.tagger.main import \
    VersionTagger, \
    ReleaseTagger, \
    ForceVersionTagger

from tito.tagger.rheltagger import RHELTagger
from tito.tagger.zstreamtagger import zStreamTagger
from tito.tagger.cargobump import CargoBump
from tito.tagger.susetagger import SUSETagger



# Copyright (c) 2008-2009 Red Hat, Inc.
#
# This software is licensed to you under the GNU General Public License,
# version 2 (GPLv2). There is NO WARRANTY for this software, express or
# implied, including the implied warranties of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. You should have received a copy of GPLv2
# along with this software; if not, see
# http://www.gnu.org/licenses/old-licenses/gpl-2.0.txt.
#
# Red Hat trademarks are not licensed under GPLv2. No permission is
# granted to use or replicate Red Hat trademarks that are incorporated
# in this software or its documentation.
"""
Tito's Command Line Interface
"""

import sys
import os
import errno

from optparse import OptionParser, SUPPRESS_HELP

from tito import __version__
from tito.common import find_git_root, error_out, debug, get_class_by_name, \
    DEFAULT_BUILDER, BUILDCONFIG_SECTION, DEFAULT_TAGGER, \
    create_builder, get_project_name, get_relative_project_dir, \
    DEFAULT_BUILD_DIR, run_command, tito_config_dir, warn_out, info_out, \
    read_user_config
from tito.compat import RawConfigParser, getstatusoutput, getoutput
from tito.exception import TitoException

# Hack for Python 2.4, seems to require we import these so they get compiled
# before we try to dynamically import them based on a string name.
import tito.tagger  # NOQA

PROGNAME = "tito"
TITO_PROPS = "tito.props"
RELEASERS_CONF_FILENAME = "releasers.conf"
ASSUMED_NO_TAR_GZ_PROPS = """
[buildconfig]
builder = tito.builder.NoTgzBuilder
tagger = tito.tagger.ReleaseTagger
"""


class FauxConfigFile(object):
    """ Allows us to read config from a string. """
    def __init__(self, config_str):
        # We'll re-add the newline when returned:
        self.lines = config_str.split("\n")

    def readline(self):
        if len(self.lines) > 0:
            # Pop a line off the front of the list:
            line = self.lines[0]
            self.lines = self.lines[1:]
            return line + "\n"
        else:
            # Indicates end of file:
            return ''


class ConfigLoader(object):
    """
    Responsible for the sometimes complicated process of loading the repo's
    tito.props, and overriding it with package specific tito.props, sometimes
    from a past tag to ensure build consistency.
    """

    def __init__(self, package_name, output_dir, tag):
        self.package_name = package_name
        self.output_dir = output_dir
        self.tag = tag

    def load(self):
        self.config = self._read_config()
        self._read_project_config()
        self._check_required_config(self.config)
        return self.config

    @property
    def tito_props_path(self):
        """
        Return absolute path to the `tito.props` file
        """
        rel_eng_dir = os.path.join(find_git_root(), tito_config_dir())
        return os.path.join(rel_eng_dir, TITO_PROPS)

    def _read_config(self):
        """
        Read global build.py configuration from the .tito dir of the git
        repository we're being run from.

        NOTE: We always load the latest config file, not tito.props as it
        was for the tag being operated on.
        """
        # Load the global config. Later, when we know what tag/package we're
        # building, we may also load that and potentially override some global
        # settings.
        config = RawConfigParser()
        config.read(self.tito_props_path)

        self._check_legacy_globalconfig(config)
        return config

    def _check_legacy_globalconfig(self, config):
        # globalconfig renamed to buildconfig for better overriding in per-package
        # tito.props. If we see globalconfig, automatically rename it after
        # loading and warn the user.
        if config.has_section('globalconfig'):
            if not config.has_section('buildconfig'):
                config.add_section('buildconfig')
            warn_out("Please rename [globalconfig] to [buildconfig] in "
                "tito.props")
            for k, v in config.items('globalconfig'):
                if k == 'default_builder':
                    warn_out("please rename 'default_builder' to "
                        "'builder' in tito.props")
                    config.set('buildconfig', 'builder', v)
                elif k == 'default_tagger':
                    warn_out("please rename 'default_tagger' to "
                        "'tagger' in tito.props")
                    config.set('buildconfig', 'tagger', v)
                else:
                    config.set('buildconfig', k, v)
            config.remove_section('globalconfig')

    def _check_required_config(self, config):
        # Verify the config contains what we need from it:
        required_global_config = [
            (BUILDCONFIG_SECTION, DEFAULT_BUILDER),
            (BUILDCONFIG_SECTION, DEFAULT_TAGGER),
        ]
        for section, option in required_global_config:
            if not config.has_section(section) or not \
                config.has_option(section, option):
                    error_out("tito.props missing required config: %s %s" % (
                        section, option))

    def _read_project_config(self):
        """
        Read project specific tito config if it exists.

        If no tag is specified we use tito.props from the current HEAD.
        If a tag is specified, we try to load a tito.props from that
        tag.
        """
        debug("Determined package name to be: %s" % self.package_name)

        # Use the properties file in the current project directory, if it
        # exists:
        current_props_file = os.path.join(os.getcwd(), TITO_PROPS)
        if (os.path.exists(current_props_file)):
            self.config.read(current_props_file)
            print("Loaded package specific tito.props overrides")

        # Check for a tito.props back when this tag was created and use it
        # instead. (if it exists)
        if self.tag:
            relative_dir = get_relative_project_dir(self.package_name, self.tag)
            debug("Relative project dir: %s" % relative_dir)

            cmd = "git show %s:%s%s" % (self.tag, relative_dir,
                    TITO_PROPS)
            debug(cmd)
            (status, output) = getstatusoutput(cmd)

            if status == 0:
                faux_config_file = FauxConfigFile(output)
                self.config.read_fp(faux_config_file)
                print("Loaded package specific tito.props overrides from %s" %
                    self.tag)
                return

        debug("Unable to locate package specific config for this package.")


def lookup_build_dir(user_config):
    """
    Read build_dir user config if it exists, otherwise
    return the current working directory.
    """
    build_dir = DEFAULT_BUILD_DIR

    if 'RPMBUILD_BASEDIR' in user_config:
        build_dir = user_config["RPMBUILD_BASEDIR"]

    return build_dir


class CLI(object):
    """
    Parent command line interface class.

    Simply delegated to sub-modules which group appropriate command line
    options together.
    """

    def main(self, argv):
        if "--version" in sys.argv:
            print(" ".join([PROGNAME, __version__]))
            sys.exit(0)

        if len(argv) < 1 or not argv[0] in CLI_MODULES.keys():
            self._usage()
            sys.exit(1)

        module_class = CLI_MODULES[argv[0]]
        module = module_class()
        return module.main(argv)

    def _usage(self):
        print("Usage: tito MODULENAME --help")
        print("Supported modules:")
        print("   build    - Build packages.")
        print("   init     - Initialize directory for use by tito.")
        print("   release  - Build and release to yum repos")
        print("   report   - Display various reports on the repo.")
        print("   tag      - Tag package releases.")


class BaseCliModule(object):
    """ Common code used amongst all CLI modules. """

    def __init__(self, usage):
        self.parser = OptionParser(usage)
        self.config = None
        self.options = None
        self.user_config = read_user_config()

        self._add_common_options()

    def _add_common_options(self):
        """
        Add options to the command line parser which are relevant to all
        modules.
        """
        # Options used for many different activities:
        self.parser.add_option("--debug", dest="debug", action="store_true",
                help="print debug messages", default=False)
        self.parser.add_option("--offline", dest="offline",
            action="store_true",
            help="do not attempt any remote communication (avoid using " +
                "this please)",
            default=False)

        default_output_dir = lookup_build_dir(self.user_config)
        self.parser.add_option("-o", "--output", dest="output_dir",
                metavar="OUTPUTDIR", default=default_output_dir,
                help="Path to write temp files, tarballs and rpms to. "
                    "(default %s)"
                    % default_output_dir)

    def main(self, argv):
        (self.options, self.args) = self.parser.parse_args(argv)

        self._validate_options()

        if len(argv) < 1:
            print(self.parser.error("Must supply an argument. "
                "Try -h for help."))

        build_dir = os.path.normpath(os.path.abspath(self.options.output_dir))
        print("Creating output directory: %s" % build_dir)
        try:
            os.makedirs(build_dir)
        except OSError as e:
            if e.errno != errno.EEXIST:
                raise

    @property
    def initial_config(self):
        """
        Configuration that is dumped into `.tito/tito.props`
        """
        config = RawConfigParser()
        config.setdefault(BUILDCONFIG_SECTION, {
            "builder": "tito.builder.Builder",
            "tagger": "tito.tagger.VersionTagger",
            "changelog_do_not_remove_cherrypick": 0,
            "changelog_format": "%s (%ae)",
        })
        return config

    def _check_config_and_load(self, loader):
        path = loader.tito_props_path
        if os.path.exists(path):
            self.config = loader.load()
            return

        # When working in non-initialized project, only building is allowed
        if not isinstance(self, BuildModule):
            error_out("Unable to locate branch configuration: %s\n"
                      "Please run 'tito init'" % loader.tito_props_path)

        if not self.options.ignore_missing_config:
            warn_out("Tito project wasn't initialized, "
                     "using the default configuration")
            warn_out("Consider `tito init' or the `--ignore-missing-config' "
                     "parameter to silent the warnings")

        self.config = self.initial_config


    def load_config(self, package_name, build_dir, tag):
        loader = ConfigLoader(package_name, build_dir, tag)
        self._check_config_and_load(loader)

        if self.config.has_option(BUILDCONFIG_SECTION,
                "offline"):
            self.options.offline = True

        if self.config.has_option(BUILDCONFIG_SECTION, "fetch_sources"):
            self.options.fetch_sources = self.config.get(BUILDCONFIG_SECTION, "fetch_sources")

        # TODO: Not ideal:
        if self.options.debug:
            os.environ['DEBUG'] = "true"

        # Check if config defines a custom lib dir, if so we add it
        # to the python path allowing users to specify custom builders/taggers
        # in their config:
        if self.config.has_option(BUILDCONFIG_SECTION,
                "lib_dir"):
            lib_dir = self.config.get(BUILDCONFIG_SECTION,
                    "lib_dir")
            if lib_dir[0] != '/':
                # Looks like a relative path, assume from the git root:
                lib_dir = os.path.join(find_git_root(), lib_dir)

            if os.path.exists(lib_dir):
                sys.path.append(lib_dir)
                debug("Added lib dir to PYTHONPATH: %s" % lib_dir)
            else:
                warn_out("lib_dir specified but does not exist: %s" % lib_dir)

    def _validate_options(self):
        """
        Subclasses can implement if they need to check for any
        incompatible cmd line options.
        """
        pass


class BuildModule(BaseCliModule):

    def __init__(self):
        BaseCliModule.__init__(self, "usage: %prog build [options]")

        self.parser.add_option("--tgz", dest="tgz", action="store_true",
                help="Build .tar.gz")
        self.parser.add_option("--srpm", dest="srpm", action="store_true",
                help="Build srpm")
        self.parser.add_option("--rpm", dest="rpm", action="store_true",
                help="Build rpm")
        self.parser.add_option("-i", "--install", dest="auto_install",
                action="store_true", default=False,
                help="Install any binary rpms being built. (WARNING: " +
                    "uses sudo rpm -Uvh --force)")
        self.parser.add_option("--no-sudo", dest="escalate",
                action="store_false", default=True,
                help="Don't escalate privileges when installing. Use when " +
                "running this command with required privileges.")
        self.parser.add_option("--dist", dest="dist", metavar="DISTTAG",
                help="Dist tag to apply to srpm and/or rpm. (i.e. .el5)")

        self.parser.add_option("--test", dest="test", action="store_true",
                help="use current branch HEAD instead of latest package tag")
        self.parser.add_option("--ignore-missing-config", action="store_true",
                help=("Acknowledge working in non-initialized project "
                      "and silencing all related warnings"))
        self.parser.add_option("--no-cleanup", dest="no_cleanup",
                action="store_true",
                help="do not clean up temporary tito build directories/files, and disable rpmbuild %clean")
        self.parser.add_option("--tag", dest="tag", metavar="PKGTAG",
                help="build a specific tag instead of the latest version " +
                    "(i.e. spacewalk-java-0.4.0-1)")

        self.parser.add_option("--builder", dest="builder",
                help="Override the normal builder by specifying a full class "
                    "path or one of the pre-configured shortcuts.")

        self.parser.add_option("--arg", dest="builder_args",
                action="append",
                help="Custom arguments specific to a particular builder."
                    " (key=value)")

        self.parser.add_option("--quiet", dest="quiet", action="store_true",
                help="Suppress output from the build process.")
        self.parser.add_option("--verbose", dest="verbose", action="store_true",
                help="Expose more output from the build process.")

        self.parser.add_option("--rpmbuild-options", dest='rpmbuild_options',
                default='',
                metavar="OPTIONS", help="Options to pass to rpmbuild.")
        self.parser.add_option("--scl", dest='scl',
                default='',
                metavar="COLLECTION", help="Build package for software collection.")

        self.parser.add_option("--fetch-sources", dest='fetch_sources',
                               action="store_true",
                               help="Download sources from predefined Source<N> addresses to the SOURCE folder")

    def main(self, argv):
        BaseCliModule.main(self, argv)

        build_dir = os.path.normpath(os.path.abspath(self.options.output_dir))
        package_name = get_project_name(tag=self.options.tag)

        build_tag = self.options.tag

        self.load_config(package_name, build_dir, self.options.tag)

        args = self._parse_builder_args()
        kwargs = {
            'dist': self.options.dist,
            'test': self.options.test,
            'ignore_missing_config': self.options.ignore_missing_config,
            'offline': self.options.offline,
            'auto_install': self.options.auto_install,
            'rpmbuild_options': self.options.rpmbuild_options,
            'scl': self.options.scl,
            'quiet': self.options.quiet,
            'verbose': self.options.verbose,
            'fetch_sources': self.options.fetch_sources,
        }

        builder = create_builder(package_name, build_tag,
                self.config,
                build_dir, self.user_config, args,
                builder_class=self.options.builder, **kwargs)
        return builder.run(self.options)

    def _validate_options(self):
        if not any([self.options.rpm, self.options.srpm, self.options.tgz]):
            error_out("Need an artifact type to build.  Use --rpm, --srpm, or --tgz")
        if self.options.srpm and self.options.rpm:
            error_out("Cannot combine --srpm and --rpm")
        if self.options.test and self.options.tag:
            error_out("Cannot build test version of specific tag.")
        if self.options.quiet and self.options.verbose:
            error_out("Cannot set --quiet and --verbose at the same time.")

    def _parse_builder_args(self):
        """
        Builder args are sometimes needed for builders that require runtime
        data.

        On the CLI this is specified with multiple uses of:

            --arg key=value

        This method parses any --arg's given and splits the key/value
        pairs out into a *dictionary of lists*.  If you only expect one value
        for the argument, you would use args['my_key'][0].
        """
        args = {}
        if self.options.builder_args is None:
            return args

        for arg in self.options.builder_args:
            if '=' in arg:
                key, value = arg.split("=", 1)
            else:
                # Allow no value args such as 'myscript --auto'
                key = arg
                value = ''

            args.setdefault(key, []).append(value)
        return args


class ReleaseModule(BaseCliModule):

    # Maps a releaser key (used on CLI) to the actual releaser class to use.
    # Projects can point to their own releasers in their tito.props.

    def __init__(self):
        BaseCliModule.__init__(self, "usage: %prog release [options] TARGET")

        self.parser.add_option("--no-cleanup", dest="no_cleanup",
                action="store_true",
                help="do not clean up temporary build directories/files")
        self.parser.add_option("--tag", dest="tag", metavar="PKGTAG",
                help="build a specific tag instead of the latest version " +
                    "(i.e. spacewalk-java-0.4.0-1)")

        self.parser.add_option("--dry-run", dest="dry_run",
                action="store_true", default=False,
                help="Do not actually commit/push anything during release.")

        self.parser.add_option("--all", action="store_true",
                help="Run all release targets configured.")

        self.parser.add_option("--test", action="store_true",
                help="use current branch HEAD instead of latest package tag")

        self.parser.add_option("-y", "--yes", dest="auto_accept", action="store_true",
                help="Do not require input, just accept commits and builds")

        self.parser.add_option("--all-starting-with", dest="all_starting_with",
                help="Run all release targets starting with the given string.")

        self.parser.add_option("-l", "--list", dest="list_releasers",
                action="store_true",
                help="List all configured release targets.")

        self.parser.add_option("--no-build", dest="no_build",
                action="store_true", default=False,
                help="Do not perform a build after a DistGit commit")

        self.parser.add_option("-s", "--scratch", dest="scratch",
                action="store_true",
                help="Perform a scratch build in Koji")
        self.parser.add_option("--arg", dest="builder_args",
                action="append",
                help="Custom arguments to pass to the builder."
                    " (key=value)")

    def _validate_options(self):

        if self.options.all and self.options.all_starting_with:
            error_out("Cannot combine --all and --all-starting-with.")

        if (self.options.all or self.options.all_starting_with) and \
                len(self.args) > 1:
            error_out("Cannot use explicit release targets with "
                    "--all or --all-starting-with.")

    def _read_releaser_config(self):
        """
        Read the releaser targets from .tito/releasers.conf.
        """
        rel_eng_dir = os.path.join(find_git_root(), tito_config_dir())
        filename = os.path.join(rel_eng_dir, RELEASERS_CONF_FILENAME)
        config = RawConfigParser()
        config.read(filename)
        return config

    def _legacy_builder_hack(self, releaser_config):
        """
        Support the old style koji builds when config is still in global
        tito.props, as opposed to the new releasers.conf.
        """
        # Handle koji:
        if self.config.has_section("koji") and not \
                releaser_config.has_section("koji"):
            warn_out("legacy 'koji' section in tito.props, please "
                    "consider creating a target in releasers.conf.")
            print("Simulating 'koji' release target for now.")
            releaser_config.add_section('koji')
            releaser_config.set('koji', 'releaser', 'tito.release.KojiReleaser')
            releaser_config.set('koji', 'autobuild_tags',
                    self.config.get('koji', 'autobuild_tags'))

            # TODO: find a way to get koji builds going through the new release
            # target config file, tricky as each koji tag gets it's own
            # section in tito.props. They should probably all get their own
            # target.

            # for opt in ["autobuild_tags", "disttag", "whitelist", "blacklist"]:
            #     if self.config.has_option("koji", opt):
            #         releaser_config.set('koji', opt, self.config.get(
            #             "koji", opt))

    def _print_releasers(self, releaser_config):
        print("Available release targets:")
        for section in releaser_config.sections():
            print("  %s" % section)

    def _calc_release_targets(self, releaser_config):
        targets = []
        if self.options.all_starting_with:
            for target in releaser_config.sections():
                if target.startswith(self.options.all_starting_with):
                    targets.append(target)
        elif self.options.all:
            for target in releaser_config.sections():
                targets.append(target)
        else:
            targets = self.args[1:]
        return targets

    def main(self, argv):
        BaseCliModule.main(self, argv)

        releaser_config = self._read_releaser_config()

        if self.options.list_releasers:
            self._print_releasers(releaser_config)
            sys.exit(1)

        # First arg is sub-command 'release', the rest should be our release
        # targets:
        if len(self.args) < 2 and (self.options.all_starting_with is None) and \
                (self.options.all is None):
            error_out("You must supply at least one release target.")

        build_dir = os.path.normpath(os.path.abspath(self.options.output_dir))
        package_name = get_project_name(tag=self.options.tag)

        self.load_config(package_name, build_dir, self.options.tag)
        self._legacy_builder_hack(releaser_config)

        targets = self._calc_release_targets(releaser_config)
        print("Will release to the following targets: %s" % ", ".join(targets))

        orig_cwd = os.getcwd()

        # Create an instance of the releaser we intend to use:
        for target in targets:
            print("Releasing to target: %s" % target)
            if not releaser_config.has_section(target):
                error_out("No such releaser configured: %s" % target)
            releaser_class = get_class_by_name(releaser_config.get(target, "releaser"))
            debug("Using releaser class: %s" % releaser_class)

            builder_args = {}
            if self.options.builder_args and len(self.options.builder_args) > 0:
                for arg in self.options.builder_args:
                    if '=' in arg:
                        key, value = arg.split("=", 1)
                    else:
                        # Allow no value args such as 'myscript --auto'
                        key = arg
                        value = ''

                    debug("Passing builder arg: %s = %s" % (key, value))
                    builder_args.setdefault(key, []).append(value)
            kwargs = {
                'builder_args': builder_args,
                'offline': self.options.offline
            }

            releaser = releaser_class(
                name=package_name,
                tag=self.options.tag,
                build_dir=build_dir,
                config=self.config,
                user_config=self.user_config,
                target=target,
                releaser_config=releaser_config,
                no_cleanup=self.options.no_cleanup,
                test=self.options.test,
                auto_accept=self.options.auto_accept,
                **kwargs)

            try:
                try:
                    releaser.release(dry_run=self.options.dry_run,
                            no_build=self.options.no_build,
                            scratch=self.options.scratch)
                except KeyboardInterrupt:
                    print("Interrupted, cleaning up...")
            finally:
                releaser.cleanup()

            # Make sure we go back to where we started, otherwise multiple
            # builders gets very confused:
            os.chdir(orig_cwd)
            print


class TagModule(BaseCliModule):

    def __init__(self):
        BaseCliModule.__init__(self, "usage: %prog tag [options]")

        # Options for tagging new package releases:
        # NOTE: deprecated and no longer needed:
        self.parser.add_option("--tag-release", dest="tag_release",
                action="store_true",
                help=SUPPRESS_HELP)
        self.parser.add_option("--keep-version", dest="keep_version",
                action="store_true",
                help=("Use spec file version/release exactly as "
                    "specified in spec file to tag package."))
        self.parser.add_option("--use-version", dest="use_version",
                help=("Update the spec file with the specified version."))
        self.parser.add_option("--use-release", dest="use_release",
                help=("Update the spec file with the specified release."))

        self.parser.add_option("--no-auto-changelog", action="store_true",
                default=False,
                help=("Don't automatically create a changelog "
                    "entry for this tag if none is found"))
        self.parser.add_option("--accept-auto-changelog", action="store_true",
                default=False,
                help=("Automatically accept the generated changelog."))

        self.parser.add_option("--auto-changelog-message",
                dest="auto_changelog_msg", metavar="MESSAGE",
                help=("Use MESSAGE as the default changelog message for "
                      "new packages"))

        self.parser.add_option("--changelog",
                dest="changelog", action="append",
                help=("Supply a custom changelog message to be used for this tag"))

        self.parser.add_option("--undo", "-u", dest="undo", action="store_true",
                help="Undo the most recent (un-pushed) tag.")

    def main(self, argv):
        BaseCliModule.main(self, argv)

        build_dir = os.path.normpath(os.path.abspath(self.options.output_dir))
        package_name = get_project_name(tag=None)

        self.load_config(package_name, build_dir, None)
        if self.config.has_option(BUILDCONFIG_SECTION,
                "block_tagging"):
            debug("block_tagging defined in tito.props")
            error_out("Tagging has been disabled in this git branch.")

        tagger_class = get_class_by_name(self.config.get(
            BUILDCONFIG_SECTION, DEFAULT_TAGGER))
        debug("Using tagger class: %s" % tagger_class)

        tagger = tagger_class(config=self.config,
                user_config=self.user_config,
                keep_version=self.options.keep_version,
                offline=self.options.offline)

        try:
            return tagger.run(self.options)
        except TitoException:
            e = sys.exc_info()[1]
            error_out(e.message)

    def _validate_options(self):
        if self.options.keep_version and self.options.use_version:
            error_out("Cannot combine --keep-version and --use-version")


class InitModule(BaseCliModule):
    """ CLI Module for initializing a project for use with tito. """

    def __init__(self):
        BaseCliModule.__init__(self, "usage: %prog init [options]")

    def main(self, argv):
        # DO NOT CALL BaseCliModule.main(self)
        # we are initializing tito to work in this module and
        # calling main will result in a configuration error.
        (self.options, self.args) = self.parser.parse_args(argv)
        should_commit = False

        rel_eng_dir = os.path.join(find_git_root(), '.tito')
        if not os.path.exists(rel_eng_dir):
            print("Creating tito metadata in: %s" % rel_eng_dir)
            os.makedirs(rel_eng_dir)
            print("   - created %s" % rel_eng_dir)
        else:
            print("Reinitializing existing tito metadata in %s" % rel_eng_dir)

        propsfile = os.path.join(rel_eng_dir, TITO_PROPS)
        if not os.path.exists(propsfile):
            with open(propsfile, 'w') as fp:
                self.initial_config.write(fp)

            print("   - wrote %s" % TITO_PROPS)

            getoutput('git add %s' % propsfile)
            should_commit = True

        # prep the packages metadata directory
        pkg_dir = os.path.join(rel_eng_dir, "packages")
        readme = os.path.join(pkg_dir, '.readme')

        if not os.path.exists(readme):
            if not os.path.exists(pkg_dir):
                os.makedirs(pkg_dir)
                print("   - created %s" % pkg_dir)

            # write out readme file explaining what pkg_dir is for
            readme = os.path.join(pkg_dir, '.readme')
            out_f = open(readme, 'w')
            out_f.write("the .tito/packages directory contains metadata files\n")
            out_f.write("named after their packages. Each file has the latest tagged\n")
            out_f.write("version and the project's relative directory.\n")
            out_f.close()
            print("   - wrote %s" % readme)

            getoutput('git add %s' % readme)
            should_commit = True

        if should_commit:
            getoutput('git commit -m "Initialized to use tito. "')
            print("   - committed to git")

        info_out("Done!")
        return []


class ReportModule(BaseCliModule):
    """ CLI Module For Various Reports. """

    def __init__(self):
        BaseCliModule.__init__(self, "usage: %prog report [options]")

        self.parser.add_option("--untagged-diffs", dest="untagged_report",
                action="store_true",
                help="%s %s %s" % (
                    "Print out diffs for all packages with changes between",
                    "their most recent tag and HEAD. Useful for determining",
                    "which packages are in need of a re-tag.",
                ))
        self.parser.add_option("--untagged-commits", dest="untagged_commits",
                action="store_true",
                help="%s %s %s" % (
                    "Print out the list for all packages with changes between",
                    "their most recent tag and HEAD. Useful for determining",
                    "which packages are in need of a re-tag.",
                ))

    def main(self, argv):
        BaseCliModule.main(self, argv)

        if self.options.untagged_report:
            self._run_untagged_report(self.config)
            sys.exit(1)

        if self.options.untagged_commits:
            self._run_untagged_commits(self.config)
            sys.exit(1)
        return []

    def _run_untagged_commits(self, config):
        """
        Display a report of all packages with differences between HEAD and
        their most recent tag, as well as a patch for that diff. Used to
        determine which packages are in need of a rebuild.
        """
        print("Scanning for packages that may need to be tagged...")
        print("")
        git_root = find_git_root()
        rel_eng_dir = os.path.join(git_root, tito_config_dir())
        os.chdir(git_root)
        package_metadata_dir = os.path.join(rel_eng_dir, "packages")
        for root, dirs, files in os.walk(package_metadata_dir):
            for md_file in files:
                if md_file[0] == '.':
                    continue
                f = open(os.path.join(package_metadata_dir, md_file))
                (version, relative_dir) = f.readline().strip().split(" ")

                # Hack for single project git repos:
                if relative_dir == '/':
                    relative_dir = ""

                project_dir = os.path.join(git_root, relative_dir)
                self._print_log(config, md_file, version, project_dir)

    def _run_untagged_report(self, config):
        """
        Display a report of all packages with differences between HEAD and
        their most recent tag, as well as a patch for that diff. Used to
        determine which packages are in need of a rebuild.
        """
        print("Scanning for packages that may need to be tagged...")
        print("")
        git_root = find_git_root()
        rel_eng_dir = os.path.join(git_root, tito_config_dir())
        os.chdir(git_root)
        package_metadata_dir = os.path.join(rel_eng_dir, "packages")
        for root, dirs, files in os.walk(package_metadata_dir):
            for md_file in files:
                if md_file[0] == '.':
                    continue
                f = open(os.path.join(package_metadata_dir, md_file))
                (version, relative_dir) = f.readline().strip().split(" ")

                # Hack for single project git repos:
                if relative_dir == '/':
                    relative_dir = ""

                project_dir = os.path.join(git_root, relative_dir)
                self._print_diff(config, md_file, version, project_dir,
                        relative_dir)

    def _print_log(self, config, package_name, version, project_dir):
        """
        Print the log between the most recent package tag and HEAD, if
        necessary.
        """
        last_tag = "%s-%s" % (package_name, version)
        try:
            os.chdir(project_dir)
            patch_command = ("git log --pretty=oneline "
                "--relative %s..%s -- %s" % (last_tag, "HEAD", "."))
            output = run_command(patch_command)
            if (output):
                print("-" * (len(last_tag) + 8))
                print("%s..%s:" % (last_tag, "HEAD"))
                print(output)
        except:
            print("%s no longer exists" % project_dir)

    def _print_diff(self, config, package_name, version,
            full_project_dir, relative_project_dir):
        """
        Print a diff between the most recent package tag and HEAD, if
        necessary.
        """
        last_tag = "%s-%s" % (package_name, version)
        os.chdir(full_project_dir)
        patch_command = "git diff --relative %s..%s" % \
                (last_tag, "HEAD")
        output = run_command(patch_command)

        # If the diff contains 1 line then there is no diff:
        linecount = len(output.split("\n"))
        if linecount == 1:
            return

        name_and_version = "%s   %s" % (package_name, relative_project_dir)
        # Otherwise, print out info on the diff for this package:
        print("#" * len(name_and_version))
        print(name_and_version)
        print("#" * len(name_and_version))
        print("")
        print(patch_command)
        print("")
        print(output)
        print("")
        print("")
        print("")
        print("")
        print("")


CLI_MODULES = {
    "build": BuildModule,
    "tag": TagModule,
    "release": ReleaseModule,
    "report": ReportModule,
    "init": InitModule,
}


def main():
    """Command line's entry point."""
    try:
        CLI().main(sys.argv[1:])
    except TitoException:
        e = sys.exc_info()[1]
        error_out(e.message)
    except KeyboardInterrupt:
        pass


if __name__ == "__main__":
    main()



#!/usr/bin/env python
#
# Copyright (c) 2008-2009 Red Hat, Inc.
#
# This software is licensed to you under the GNU General Public License,
# version 2 (GPLv2). There is NO WARRANTY for this software, express or
# implied, including the implied warranties of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. You should have received a copy of GPLv2
# along with this software; if not, see
# http://www.gnu.org/licenses/old-licenses/gpl-2.0.txt.
#
# Red Hat trademarks are not licensed under GPLv2. No permission is
# granted to use or replicate Red Hat trademarks that are incorporated
# in this software or its documentation.
"""
Tito Setup Script
"""

from io import open
from setuptools import setup, find_packages


with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()


setup(
    name="tito",
    version='0.6.26',
    description='A tool for managing rpm based git projects.',
    long_description=long_description,
    long_description_content_type="text/markdown",
    author='rpm-software-management',
    url='https://github.com/rpm-software-management/tito',
    license='GPLv2',

    # tell distutils packages are under src directory
    package_dir={
        '': 'src',
    },
    packages=find_packages('src'),
    include_package_data=True,
    install_requires=[
        'blessed'
    ],

    # automatically create console scripts
    entry_points={
        'console_scripts': [
            'tito = tito.cli:main',
        ],
    },

    # non-python scripts go here
    scripts=[
        'bin/generate-patches.pl'
    ],

    classifiers=[
        'License :: OSI Approved :: GNU General Public License (GPL)',
        'Development Status :: 2 - Pre-Alpha',
        'Environment :: Console',
        'Intended Audience :: Developers',
        'Intended Audience :: Information Technology',
        'Programming Language :: Python'
    ],
)


# XXX: this will also print on non-install targets
print("tito target is complete")
